#!/usr/bin/perl -w

# Linux Application Checker
# Results Page Module (Tests_results.pm)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Tests_results;

use Templates;
use Analysis; # utils/Analysis.pm

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	&read_results_info
	&html_title &html_results_tabs
	&load_analysis_data
	&get_distr_sel
	&Summary
    &D_tab
		&distr_details
	&C_tab
		&com_details &com_common_details &com_distr_details
	&L_tab
		&lib_details &lib_distr_list
	&I_tab
		&int_details &int_distr_list
);

our @EXPORT_OK = qw(&tab_link &href &html);

# Colorscheme:
#  #CCFFCC - green background
#  #FEFECC - yellow background
#  #FEECCC - dark yellow background
#  #FFCCCC - dark red background
#  #FFDDDD - red background
#
#  'red'   - red foreground
#  #A04020 - yellow-red foreground
#  #907000 - yellow foreground
#  'green' - green foreground
#  'blue'  - blue foreground

#-----------------------------------------------------------------------
# Misc stuff:

sub tab_link ($) {
	my ($tab) = @_;
	
	return '?details='.$_GET{'details'}.'&tab='.$tab;
}

sub href ($) {
	my ($link) = @_;
	
	$link =~ s/\+/%2B/g;
	$link =~ s/ /%20/g;
	
	return $link;
}

sub html ($) {
	my ($txt) = @_;
	
	$txt =~ s/&/&amp;/g;
	$txt =~ s/</&lt;/g;
	$txt =~ s/>/&gt;/g;
	$txt =~ s/\"/&quot;/g;
	
	return $txt;
}
#-----------------------------------------------------------------------

# Reads information about the tested application
sub read_results_info($) {
	my ($res_dir) = @_;

	my $app_name = "";
	if (open(FILE, "$res_dir/INFO")) {
		while (<FILE>) {
			if (m/Application: (.*)$/) {
				$app_name = $1;
				last;
			}
		}
		close(FILE);
	}
	my $res_dir_esc = $res_dir;
	$res_dir_esc =~ s/([ ()\'])/\\$1/g;
	my $host_platform = `cat $res_dir_esc/test_logs/host_platform`;
	my $lsb_version = `cat $res_dir_esc/test_logs/lsb_version`;
	$host_platform =~ s/[\r\n]//g;
	$lsb_version =~ s/[\r\n]//g;

	return ($app_name, $host_platform, $lsb_version);
}
#-----------------------------------------------------------------------

# Renders main title of the results page.
sub html_title {
	my ($app_name, $architecture, $res_dir) = @_;

	my $res = '';
	if (! -f "$res_dir/test_logs/ftp_upload") {
		my $href_name = $_GET{'details'};
		$href_name =~ s/([^0-9a-zA-Z_-])/sprintf('%%%02X', ord($1))/eg;
		$res .= <<DATA;
<table border="0" cellpadding="0" cellspacing="0" class="upload-suggest" onclick="javascript:document.location='upload_data.pl?details=$href_name';">
  <tr>
    <td valign="middle"><img src="images/upload_data.png" width="16" height="16" alt="" /></td>
    <td><img src="images/environment-spacer.gif" width="3" height="1" alt="" /></td>
    <td>Please <a href="upload_data.pl?details=$href_name" onclick="javascript:event.cancelBubble=true;"><b>upload</b></a> info about your application to the <a href="http://www.linuxfoundation.org/navigator/browse/app.php" target="_blank" onclick="javascript:event.cancelBubble=true;">LSB Navigator</a>.</td>
  </tr>
</table>
<script type="text/javascript" language="javascript">
// <![CDATA[
preload_images(new Array('ajax_progress.gif'));
// ]]>
</script>
DATA
	$res .= "<h1>Analysis Results for <font color=\"blue\">$app_name</font> on <font color=\"blue\">$architecture</font></h1>\n\n";

	}
	return $res;
}
#-----------------------------------------------------------------------

my @res_tab_ids = ('distrs', 'com', 'libs', 'ints', 'cert');
my %res_tab_names = (
	'distrs' => 'Distribution Compatibility',
	'com'    => 'App Components',
	'libs'   => 'External Libraries',
	'ints'   => 'External Interfaces',
	'cert'   => 'LSB Certification',
);

# Customizable tab bar
sub html_results_tabs {
	my ($selected) = @_;

	my $nl = "\n";
	my $res = "";

	# Table start
	$res .= '<table cellpadding="0" cellspacing="0" border="0" width="100%"'
		.' style="white-space: nowrap; font-weight: bold; text-align: center; font-size: 120%; margin-bottom: 5pt;">'.$nl;
	$res .= '<tr>'.$nl;

	# Left spacer
	$res .= '<td style="width: 10px; border-bottom: solid 1px #166a98; height: 2em;">';
	$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
	$res .= '</td>'.$nl;

	my $tabs_num = scalar(@res_tab_ids);

	foreach my $id (@res_tab_ids) {
		my $name = $res_tab_names{$id};

		# Tab
		if ( $id eq $selected ) {
			$res .= '<td class="res-tab-selected">';
		} else {
			$res .= '<td class="res-tab-highlight">';
			$res .= '<a class="res-tab-link" href="'.html(href(tab_link($id))).'">';
		}
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '<img src="images/res-'.$id.'.png" width="16" height="16" alt="" />&nbsp;';
		$name =~ s/\s/&nbsp;/g;
		$res .= $name;
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '</a>' if $id ne $selected;
		$res .= '</td>';

		# Spacer
		$res .= '<td style="width: 10px; border-bottom: solid 1px #166a98;">';
		$res .= '<img src="images/environment-spacer.gif" width="10" height="1" alt="" />';
		$res .= '</td>'.$nl;
	}

	$res .= '<td style="border-bottom: solid 1px #166a98; width: 100%;">';
	$res .= '<img src="images/environment-spacer.gif" width="1" height="1" alt="" />';
	$res .= '</td>';

	$res .= '</tr>';

	# Bottom line
	$res .= '<tr>';
	$res .= '<td colspan="'.(2 * $tabs_num+2).'" style="background-color: #a0d5f1; height: 10px;">';
	$res .= '<img src="images/environment-spacer.gif" width="1" height="1" alt="" />';
	$res .= '</td>';
	$res .= '</tr>';

	$res .= '</table>'.$nl;

	return $res;
}
#-----------------------------------------------------------------------

# Returns distribution selection passed by POST.
# Struct: $distr_sel->{$Did} = $checked;
sub get_distr_sel {
	my $distr_sel = {};
	
	if ( $_COOKIE{"ds"} ) {
		# User's selection
		for my $Did ( split /-/, $_COOKIE{"ds"} ) {
			next if !$Analysis::all_distrs->{$Did};
			$distr_sel->{$Did} = 1;
		}
		for my $Did ( keys %$Analysis::all_distrs ) {
			if ( !$distr_sel->{$Did} ) {
				$distr_sel->{$Did} = 0;
			}
		}
	} else {
		# Default values
		for my $Did ( keys %$Analysis::all_distrs ) {
			$distr_sel->{$Did} = 1;
		}
	}
	
	return $distr_sel;
}
#-----------------------------------------------------------------------

# Uses Analysis module to load all the analysis results.
sub load_analysis_data {
	my ($res_dir) = @_;
	
	# Load distribution list (contains distribution name, version, vendor, release year)
	Analysis::load_distrs( $res_dir."/Distributions" )
		or return "Failed to read file '$res_dir/Distributions'.<br/>This may happen if the Analyzer failed.";
	
	my $err = Analysis::load_main( $res_dir."/analysis_data" );
	if ( $err ) {
		return html($err)."<br/>Failed to read file '$res_dir/analysis_data'.";
	}
	
	Analysis::complete_structures()
		or return "Mailformed data was loaded.";
	
	Analysis::unneeded_libs();
	
	return;
}
#-----------------------------------------------------------------------

# Renders a control for selecting distribution subset
sub distr_selector {
	my ($distr_sel) = @_;

	my $nl = "\n";
	my $res = '';

	my $sel_distrs = scalar grep {$_} values %$distr_sel;
	my $total_distrs = scalar keys %$distr_sel;

	my %distrs = ();
	foreach my $did (keys %$distr_sel) {
		my $distr = $Analysis::all_distrs->{$did};
		$distrs{$did} = {
			'NAME'   => $distr->{'FULLNAME'},
			'YEAR'   => $distr->{'YEAR'},
			'FAMILY' => $distr->{'NAME'},
			'VENDOR' => $distr->{'VENDOR'}
		};
	}

	my @groups = (
		{
			'NAME'        => 'no',
			'DESCRIPTION' => '* No grouping *',
			'SORT_FUNC'   => undef,
			'ICON_FUNC'   => undef
		},
		{
			'NAME'        => 'FAMILY',
			'DESCRIPTION' => 'By distro family',
			'SORT_FUNC'   => sub {lc($a) cmp lc($b)},
			'ICON_FUNC'   => sub($) {'images/distros/'.lc($_[0]).'.png'}
		},
		{
			'NAME'        => 'VENDOR',
			'DESCRIPTION' => 'By vendor',
			'SORT_FUNC'   => sub {lc($a) cmp lc($b)},
			'ICON_FUNC'   => sub($) {'images/vendors/'.lc($_[0]).'.png'}
		},
		{
			'NAME'        => 'YEAR',
			'DESCRIPTION' => 'By release year',
			'SORT_FUNC'   => sub {$b <=> $a},
			'ICON_FUNC'   => undef
		}
	);
	my $num_groups = scalar(@groups);

	my @distr_groups = ();
	my @group_sel = ();
	foreach (my $i = 1; $i < $num_groups; ++$i) {
		my $group = $groups[$i];
		my $gname = $group->{'NAME'};
		$distr_groups[$i] = {};
		$group_sel[$i] = {};
		foreach my $did (keys %distrs) {
			my $group_value = $distrs{$did}->{$gname};
			if (defined($group_sel[$i]->{$group_value})) {
				if (!$distr_sel->{$did}) {
					$group_sel[$i]->{$group_value} = 0;
				}
			}
			else {
				$group_sel[$i]->{$group_value} = $distr_sel->{$did};
			}
			if (!defined($distr_groups[$i]->{$group_value})) {
				$distr_groups[$i]->{$group_value} = [];
			}
			push @{$distr_groups[$i]->{$group_value}}, $did;
		}
	}

	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
var num_distrs = $sel_distrs;
var num_groups = $num_groups;

var distros_checked = new Array();
var distro_to_group = new Array();
var group_to_distros = new Array();

HEREDOC

	foreach my $did (sort {$a <=> $b} keys %$distr_sel) {
		$res .= "distros_checked[$did] = $distr_sel->{$did};\n";
	}
	$res .= "\n";

	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= "// Group $i: $groups[$i]->{'NAME'}\ndistro_to_group[$i] = new Array();\n";
		foreach my $did (sort {$a <=> $b} keys %distrs) {
			my $group_value = $distrs{$did}->{$groups[$i]->{'NAME'}};
			$group_value =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			$res .= "distro_to_group[$i][$did] = '$group_value';\n";
		}
	}
	$res .= "\n";

	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= "// Group $i: $groups[$i]->{'NAME'}\ngroup_to_distros[$i] = new Array();\n";
		foreach my $group_value (sort {lc $a cmp lc $b} keys %{$distr_groups[$i]}) {
			my $arr = join(', ', @{$distr_groups[$i]->{$group_value}});
			$group_value =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			$res .= "group_to_distros[$i]['$group_value'] = [$arr];\n";
		}
	}
	$res .= "\n";

	my $cur_ds = join "-", sort {$a<=>$b} grep {$distr_sel->{$_}} keys %$distr_sel;

	$res .= <<HEREDOC;
function update_sel_distrs() {
	var sel_distrs = document.getElementById('sel-distrs');
	if ( sel_distrs ) {
		sel_distrs.innerHTML = num_distrs;
		if ( num_distrs < $total_distrs ) {
			sel_distrs.style.color = 'red';
		} else {
			sel_distrs.style.color = '';
		}
	}
}

function my_reset() {
	var inputs = document.getElementsByTagName('input');
	for (var i = 0; i < inputs.length; i++) {
		if ((inputs[i].type == 'checkbox') && ((inputs[i].id.indexOf('distr-') == 0) || (inputs[i].id.indexOf('group-') == 0))){
			inputs[i].checked = 'checked';
		}
	}
	for (var did in distros_checked) {
		distros_checked[did] = 1;
	}
	num_distrs = $total_distrs;
	update_sel_distrs();
}

function checkGroupCheckbox(group_idx, group_value) {
	// Do not check the 'no' grouping
	if (group_idx == 0)
		return;
	for (var i=0; i<group_to_distros[group_idx][group_value].length; ++i) {
		var did = group_to_distros[group_idx][group_value][i];
		if (!document.getElementById('distr-' + group_idx + '-' + did).checked) {
			document.getElementById('group-' + group_idx + '-' + group_value).checked = false;
			return;
		}
	}
	document.getElementById('group-' + group_idx + '-' + group_value).checked = true;
}

function onDistrToggle(elem) {
	distrToggle(elem.id.substr(elem.id.lastIndexOf('-') + 1), elem.checked, 0);
	update_sel_distrs();
}

function distrToggle(did, checked, not_check_group) {
	if (distros_checked[did] == checked)
		return;
	distros_checked[did] = checked;
	if (checked)
		++num_distrs;
	else
		--num_distrs;
	for (var i=0; i<num_groups; ++i) {
		document.getElementById('distr-' + i + '-' + did).checked = checked;
		if ((i > 0) && (i != not_check_group))
			checkGroupCheckbox(i, distro_to_group[i][did]);
	}
}

function onGroupToggle(elem, group_idx, group_value) {
	// Do not check the 'no' grouping
	if (group_idx == 0)
		return;
	var checked = elem.checked;
	for (var i=0; i<group_to_distros[group_idx][group_value].length; ++i) {
		var did = group_to_distros[group_idx][group_value][i];
		distrToggle(did, checked, group_idx);
	}
	update_sel_distrs();
}

function switch_grouping(id) {
	for (var i=0; i<num_groups; ++i) {
		var div = document.getElementById('distr-grouping-' + i);
		if (div)
			div.style.display = ((i == id) ? '' : 'none');
	}
}

var cur_ds = '$cur_ds';

function distr_sel_apply() {
	var distrsel_form = document.getElementById('distrsel');
	if ( !distrsel_form ) return;
	
	var distrs = "";
	
	var inputs = distrsel_form.getElementsByTagName('input');
	for (var i = 0; i < inputs.length; i++) {
		if ( !inputs[i].checked ) continue;
		if ( inputs[i].name.match(/^distr-0-\\d+\$/) ) {
			var did = inputs[i].name.replace(/^distr-0-/, "");
			if ( distrs ) distrs += "-";
			distrs += did;
		}
	}
	
	cur_ds = distrs;
	setCookie('ds', distrs, null);
	
	onFilter();
}
// ]]>
</script>
HEREDOC

	# Toggler
	$res .= '<div>';
	$res .= '<a href="" onclick="javascript:return toggle(\'dseldiv\');"><b><u>Select distributions&hellip;</u></b></a> (<b><span id="sel-distrs">'.($sel_distrs==$total_distrs ? $sel_distrs : '<span style="color:red">'.$sel_distrs.'</span>').'</span></b> of <b>'.$total_distrs.'</b> selected)';
	$res .= '</div>';
	
	# Print selector
	$res .= '<div id="dseldiv" style="display: none;">';

	# Print group options
	$res .= <<DATA;
<table style="border: solid 1px gray; background: #e0e0e0; padding: 2px; margin: 0.5em 0;"><tr><td>
Group: <select name="distr-grouping" id="distr-grouping" onchange="javascript:switch_grouping(this.value);" style="width: 12.5em;">
DATA
	foreach (my $i = 0; $i < $num_groups; ++$i) {
		$res .= "  <option value=\"$i\"".(($i == 0) ? ' selected="selected"' : '').">$groups[$i]->{'DESCRIPTION'}</option>\n";
	}
	$res .= <<DATA;
</select>
</td></tr></table>
DATA

	# Print DIVs for every grouping
	for (my $i = 1; $i < $num_groups; ++$i) {
		$res .= '<div id="distr-grouping-'.$i.'" style="border: solid 1px; padding: 3pt; margin-bottom: 3pt; height: 200pt; width: 200pt; overflow: auto; display: none; white-space: nowrap;">'."\n";
		foreach my $group_value (sort {$groups[$i]->{'SORT_FUNC'}()} keys %{$distr_groups[$i]}) {
			my $group_value_id = $group_value;
			$group_value_id =~ s/([^a-zA-Z0-9_:])/sprintf('.%02X', ord($1))/eg;
			my $g_checked = ($group_sel[$i]->{$group_value} ? ' checked="checked"' : '');
			my $g_icon = (defined($groups[$i]->{'ICON_FUNC'}) ? '&nbsp;<img src="'.$groups[$i]->{'ICON_FUNC'}($group_value).'" width="16" height="16" alt="" />' : '');
			$res .= <<DATA;
  <label><input type="checkbox" name="group-$i-$group_value_id" id="group-$i-$group_value_id"$g_checked onclick="javascript:onGroupToggle(this, $i, '$group_value_id');" />$g_icon&nbsp;$group_value</label><br />
DATA
			foreach my $did (sort {lc($distrs{$a}->{'NAME'}) cmp lc($distrs{$b}->{'NAME'})} @{$distr_groups[$i]->{$group_value}}) {
				my $d_checked = ($distr_sel->{$did} ? ' checked="checked"' : '');
				my $d_icon = '&nbsp;<img src="images/distros/'.lc($distrs{$did}->{'FAMILY'}).'.png" width="16" height="16" alt="" />';
				$res .= <<DATA;
  &nbsp;&nbsp;&nbsp;&nbsp;<label><input type="checkbox" name="distr-$i-$did" id="distr-$i-$did"$d_checked onclick="javascript:onDistrToggle(this);" />$d_icon&nbsp;$distrs{$did}->{'NAME'}</label><br />
DATA
			}
		}
		$res .= "</div>\n";
	}

	# Print DIV with plain list
	$res .= <<DATA;
<form name="distrsel" id="distrsel" method="post" action="">
<div id="distr-grouping-0" style="border: solid 1px; padding: 3pt; margin-bottom: 3pt; height: 200pt; width: 200pt; overflow: auto; white-space: nowrap;">
DATA
	foreach my $did (sort {lc($distrs{$a}->{'NAME'}) cmp lc($distrs{$b}->{'NAME'})} keys %distrs) {
		my $d_checked = ($distr_sel->{$did} ? ' checked="checked"' : '');
		my $d_icon = '&nbsp;<img src="images/distros/'.lc($distrs{$did}->{'FAMILY'}).'.png" width="16" height="16" alt="" />';
		$res .= <<DATA;
  <label><input type="checkbox" name="distr-0-$did" id="distr-0-$did"$d_checked onclick="javascript:onDistrToggle(this);" />$d_icon&nbsp;$distrs{$did}->{'NAME'}</label><br />
DATA
	}
	$res .= <<DATA;
</div>
<input type="button" value="Apply" style="width: 7em;" onclick="distr_sel_apply();"/>
 &nbsp; &nbsp; 
<input type="button" value="Reset" style="width: 7em;" onclick="my_reset();" />
</form>
</div>
DATA

	# Print hidden form with selected distros
	$res .= '<form name="hiddensel" action="" method="post" style="display:none">'.$nl;
	foreach my $Did ( keys %$distr_sel ) {
		my $control_name = "distr-0-".$Did;

		$res .= '<input type="hidden" name="'.$control_name.'"'
				.' value="'.($distr_sel->{$Did}?1:0).'" />'.$nl;
	}
	$res .= '<input type="hidden" name="distrsel" value="1" />'.$nl;
	$res .= '</form>'.$nl;

	return $res;
}

sub parse_app_component {
	my ($s) = @_;
	
	my $disp_path = $s;
	$disp_path =~ s/\|//g;
	
	$s =~ s/^.*\.rpm\|([^|]+)/RPM:$1/;
	$s =~ s/^.*\.deb\|([^|]+)/DEB:$1/;
	$s =~ s/^.*\.(tar)\|([^|]+)/TAR:$2/;
	$s =~ s/^.*\.(tar\.gz|tgz)\|([^|]+)/TGZ:$2/;
	$s =~ s/^.*\.(tar\.bz2|tbz)\|([^|]+)/TBZ:$2/;
	$s =~ s/^PKG:.*\|([^|]+)/PKG:$1/;
	$s =~ s/\|//g;
	while ( length($s) > 50 ) { # shrink long pathes via ellipsis
		$s =~ s{^((\w+:)?)(...)?/[^/]+/}{$1.../} or last;
	}
	
	return ($s, $disp_path);
}

sub no_binary_div {
	my $res = "";
	
	$res .= '<div style="background:#fec;border:1px solid black;padding:10pt;margin-top:20pt;text-align:center;font-size:large">';
	$res .= $_[0] || 'This application contains no binary files (ELF executables or libraries).<br />Please see the LSB Certification tab for the results of script analysis (Perl, Python, Shell scripts).';
	$res .= '</div>';

	return $res;
}
#-----------------------------------------------------------------------

sub libs_data {
	my ($filt_distr, $filt_com) = @_;
	
	my $libs = {};
	# $libs->{$soname} => { # per column
	#  COMPONENTS->{$com_name} => 1,
	#  INTERFACES->{$fullname} => 1,
	#  DISTRS->{$Did} => 1,
	#  LSB_STATUS => {STATUS, STATUS_VER},
	#  MORE_INFO => "MORE_INFO"/"KNOWN_ISSUES",
	#  APPROVED => 1/0,
	# };
	
	my $nonfilt = { LIBS => {} }; # Non-filtered information

	foreach my $com_obj ( values %$Analysis::app_components ) {
		# Non-filtered information
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			$nonfilt->{LIBS}{$soname} = 1;
		}
		# Filter
		if ( $filt_com && $com_obj->{NAME} ne $filt_com ) { next; } # component filter
		
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			$libs->{$soname}{COMPONENTS}{$com_obj->{NAME}} = 1; # COMPONENTS field
		}
		foreach my $Did ( keys %$Analysis::all_distrs ) {
			if ($filt_distr && !$filt_distr->{$Did}) { next; } # distribution filter
			
			my $com_distr = $com_obj->{$Did};
			my $com_distr_libs = $com_distr->{LIBS};
			
			# Presence in distributions
			foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
				next if !$com_distr_libs->{$soname};
				
				if ( $com_distr_libs->{$soname}{FOUND} ) {
					unless ( defined $libs->{$soname}{DISTRS}{$Did} ) { # do not set 1 if there is already 0 or 1
						$libs->{$soname}{DISTRS}{$Did} = 1; # DISTRS
					}
				} else {
					$libs->{$soname}{DISTRS}{$Did} = 0; # DISTRS
				}
			}
			# Gather interfaces
			foreach my $fullname ( keys %{$com_distr->{FOUND_INTS}} ) {
				my $soname = $com_distr->{FOUND_INTS}{$fullname};
				$soname =~ s/\|.*//; # get soname
				
				my $lib = $com_distr_libs->{$soname}; # should be defined
				if ( $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$libs->{$soname}{INTERFACES}{$fullname} = 1;
			}
			foreach my $fullname ( keys %{$com_distr->{MISSING_INTS}} ) {
				my $soname = $com_distr->{MISSING_INTS}{$fullname};
				
				my $lib = $com_distr_libs->{$soname}; # may be undef
				if ( $lib && $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$libs->{$soname}{INTERFACES}{$fullname} = 1;
			}
		}
		# TODO: LIBS_OTHER_VERSION
	}
	
	foreach my $soname ( keys %$libs ) {
		# APPROVED
		$libs->{$soname}{APPROVED} = ($Analysis::approved_libs->{$soname} or 0);
	}
	
	if ( wantarray ) {
		return ($libs, $nonfilt);
	}
	return $libs;
}

sub libs_lsb_data {
	my ( $libs, $selected_lsb_ver ) = @_;
	
	# LSB status
	foreach my $soname ( keys %$libs ) {
		my $lib_obj = $libs->{$soname};
		
		$lib_obj->{LSB_STATUS} = Analysis::lib_lsb_status($soname, $selected_lsb_ver);
		
		if ( $Analysis::lib_attrs->{$soname} ) {
			# MORE_INFO
			$lib_obj->{MORE_INFO} = 'MORE_INFO';
		}
		
		if ( $Analysis::libKBDistr->{$soname} ) {
			# KNOWN_ISSUES
			$lib_obj->{MORE_INFO} = 'KNOWN_ISSUES';
		}
	}
	
	return undef;
}

# Library compatibility table
sub L_tab {
	my ($selected_lsb_ver, $distr_sel, $filt_com, $filt_show) = @_;

	my $res = "";
	my $nl = "\n";
	
	my $filtered = $filt_show || $filt_com || scalar grep {!$_} values %$distr_sel;
	
	my ($libs, $nonfilt_info) = libs_data( $distr_sel, $filt_com );
	libs_lsb_data($libs, $selected_lsb_ver);
	
	# Check if the table is empty
	if ( !$filtered && !%$Analysis::app_components ) {
		$res .= no_binary_div();
		return $res;
	}
	if ( !$filtered && !%$libs ) {
		$res .= no_binary_div("The application needs no external libraries.");
		return $res;
	}

	$res .= <<HEREDOC;
The table below lists all external libraries required by your Application (based on DT_NEEDED ELF section). LSB and Distribution presence statuses are provided for each library. You can adjust distributions of interest and the report will be updated appropriately.<br /><br />
HEREDOC

	my $link = href(tab_link('libs'));
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id )
{
	var v = document.getElementById( id );
	if ( v ) {
		v.style.display = (v.style.display == 'none' ? '' : 'none');
	}
	return false;
}

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes,resizable=1'
		);
	w.moveTo(100,0);
	w.focus();

	return false;
}
// ]]>
</script>
HEREDOC
	
	my ($table, $row_n) = L_table($libs, $selected_lsb_ver, $distr_sel, $filt_com, $filt_show);
	
	# Filters
	my $filt = "";
	
	$filt .= distr_selector( $distr_sel ); # Distribution selector
	
	$filt .= '<form name="filter" method="get" action="">';
	$filt .= '<table><tr><td>';
	
	$filt .= '<table cellspacing="0" class="bartable"><tr>';
	
	# "Component" filter
	$filt .= '<td style="border-right:0"><b>Component:</b></td>'
		.'<td>'
		.'<select name="filt_com" id="filt_com" onchange="onFilter()" style="width: 20em;">'.$nl
		.'<option value="">All</option>'.$nl;
	foreach my $com ( sort keys %$Analysis::app_components ) {
		my ($disp_name, $disp_path) = parse_app_component($com);
		my $selected = "";
		if ( $filt_com && $com eq $filt_com ) {
			$selected = ' selected="selected"';
		}
		$filt .= '<option value="'.html($com).'"'
					.$selected.'>'.html($disp_name).'</option>'.$nl;
	}
	$filt .= '</select>'.$nl
		.'</td>'; # </component>
	
	# spacer
	$filt .= '<td style="padding:0.2em;border:0;background:white"></td>';
	
	# "Show" filter
	$filt .= '<td style="border-right:0"><b>Show:</b></td>'
		.'<td>'
		.'<select name="filt_show" id="filt_show" onchange="onFilter()" style="width: 12.5em;">'.$nl;
	my @filt_show_values = ( ""=>"All", "problems"=>"Compatibility problems",
			"unneeded"=>"Unneeded libraries", "uncommon"=>"Uncommon libraries",
			"non-lsb"=>"Non-LSB", "deprecated"=>"Deprecated" ); # yes, this IS a LIST.
	while ( @filt_show_values ) {
		my $key = shift @filt_show_values;
		my $val = shift @filt_show_values;
		my $selected = "";
		if ( $filt_show && $key eq $filt_show ) {
			$selected = ' selected="selected"'; 
		}
		$filt .= '<option value="'.$key.'"'
					.$selected.'>'.$val.'</option>'.$nl;
	}
	$filt .= '</select>'.$nl
		.'</td>'; # </show filter>
	
	# "Apply" button
	#$filt .= '<td style="padding:0.2em;border:0;background:white"></td>';
	#$filt .= '<td onclick="return onFilter()" style="cursor:pointer"><b>Apply &gt;</b></td>';
	
	# TODO: highlight changed filters
	
	$filt .= '</tr></table>';
	$filt .= '</td>';

	# Filter status
	my $all_n = scalar keys %{$nonfilt_info->{LIBS}};
	$filt .= '<td style="padding-left:20pt">';
	$filt .= '<span id="lines_count">';
	if ( $filtered ) {
		$filt .= '<span style="color:red"><b>'.$row_n.'</b></span> of ';
	}
	$filt .= '</span>';
	$filt .= $all_n.' librar'.($all_n!=1?'ies':'y').' shown.';
	$filt .= '</td></tr></table>'; # </filters>
	$filt .= '</form>';
	
	## Ajax stuff
	$res .= show_error_dlg(''); # placeholder for an error box
	
	$res .= ajax_wait(); # ajax hourglass
	
	my $ajax_link = href('details='.$_GET{'details'});
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[

timeout_show_progress = 0;

var cur_lfc = '$filt_com';
var cur_lfs = '$filt_show';

function onFilter() {
	var filt_com = document.getElementById('filt_com');
	var filt_show = document.getElementById('filt_show');
	
	if ( filt_com ) {
		cur_lfc = filt_com.value;
		setCookie('lfc', cur_lfc, null);
	}
	if ( filt_show ) {
		cur_lfs = filt_show.value;
		setCookie('lfs', cur_lfs, null);
	}
	
	if (!ajax_req_busy) {
		ajax_call_get('action=results_table&table=libs&${ajax_link}');
	}
	
	return false;
}

function ajaxProcessResult(responseXML) {
	
	var l_table_html;
	if (responseXML.getElementsByTagName('l_table').length > 0) {
		l_table_html = responseXML.getElementsByTagName('l_table')[0].childNodes[0].nodeValue;
	}
	
	var l_table_wrap = document.getElementById('l_table_wrap');
	
	if ( l_table_html && l_table_wrap ) {
		l_table_wrap.innerHTML = l_table_html;
	} else {
		return -1;
	}
	
	var lines_count = document.getElementById('lines_count');
	if ( lines_count ) {
		var row_n;
		if (responseXML.getElementsByTagName('row_n').length > 0) {
			row_n = responseXML.getElementsByTagName('row_n')[0].childNodes[0].nodeValue;
			lines_count.innerHTML = '<span style="color:red"><b>'+row_n+'</b></span> of&nbsp;';
		} else {
			lines_count.innerHTML = '';
		}
	}
}

// Cookies on focus:
if (typeof(document.addEventListener) != 'undefined') {
 document.addEventListener('focus', doFocus, false); 
} else if (typeof(window.attachEvent) != 'undefined') {
 window.attachEvent('onfocus', doFocus);
}

function doFocus(event) {
	setCookie('ds', cur_ds, null);
	setCookie('lfc', cur_lfc, null);
	setCookie('lfs', cur_lfs, null);

}

// ]]>
</script>
HEREDOC
	
	## Put the filters
	$res .= $filt;
	
	## Put the table
	$res .= '<div id="l_table_wrap">'.$nl;
	$res .= $table;
	$res .= '</div>'.$nl;
	##
	
	return $res;
}
#-----------------------------------------------------------------------

sub L_table {
	my ($libs, $selected_lsb_ver, $distr_sel, $filt_com, $filt_show) = @_;
	
	my $filt_distr = scalar grep {!$_} values %$distr_sel;
	my $filtered = $filt_show || $filt_com || $filt_distr;
	
	my $table = "";
	my $nl = "\n";
	
	# Table Header
	$table .= '<table id="l_table" class="wikitable" width="100%">'.$nl;
	$table .= '<tr>'
		.'<th>Library</th>'
		.'<th>Dependent<br />Application Components</th>'
		.'<th>Interfaces<br />Actually Used</th>'
		.'<th>Presence in<br />'.($filt_distr ? 'Selected ' : '').'Distributions</th>'
		.'<th id="lsb_col_header">In LSB '.$selected_lsb_ver.'?</th>'
		.'<th>More Info</th>'
		.'</tr>'.$nl;
	
	my $row_n = 0; # Count the rows

	# Table Body
	foreach my $soname ( sort {lc $a cmp lc $b} keys %$libs ) {
		my $lib_obj = $libs->{$soname};

		my $row = "";
		
		$row .= '<tr>'.$nl;

		## Library
		my $navi_link = href( 'http://www.linuxfoundation.org/navigator/browse/rawlib.php'
				.'?cmd=list-byrunname&RLrunname='.$soname );
		$row .= '<td><b><a href="'.html($navi_link).'" target="_blank">'.html($soname).'</a></b></td>'.$nl;

		# App Modules: preparation
		my $app_components = "";
		my $app_comp_n = 0;
		if ( $lib_obj->{COMPONENTS} ) {
			foreach my $com ( sort keys %{$lib_obj->{COMPONENTS}} ) {
				my ($disp_name, $disp_path) = parse_app_component($com);
				
				$app_components .= '<div class="indentwrap" title="'.html($disp_path).'">'.html($disp_name).'</div>';
				
				$app_comp_n++;
			}
		}
		
		# TODO: components: 1 of 12 due to filter (need $nonfilt_data->{COMPONENTS})
		
		## App modules expandable list
		$row .= '<td>';
		my $div_id = "cl".$row_n;
		my $plu = ( $app_comp_n > 1 ? 's' : '' );
		$row .= '<a href="" onclick="return toggle(\''.$div_id.'\')">'
					.$app_comp_n.' component'.$plu.' (&hellip;)'
				.'</a>';
		$row .= '<div id="'.$div_id.'" style="display:none">';
		$row .= $app_components;
		$row .= '</div>';
		$row .= '</td>'.$nl;

		# Interfaces used
		my $ints_used = 0;
		if ( $lib_obj->{INTERFACES} ) {
			$ints_used = scalar keys %{$lib_obj->{INTERFACES}};
		}
		$row .= '<td align="center">';
		if ( $ints_used ) {
			if ( $filt_show && $filt_show eq 'unneeded' ) { next; }
			my $link = href(tab_link('ints').'&filt_lib='.$soname);
			$link .= href('&filt_com='.$filt_com) if $filt_com;
			$row .= '<a href="'.html($link).'">'.$ints_used.'</a>'; # Blue link
		} else {
			$row .= '<span style="color:red"><b>'.$ints_used.'</b></span>'; # Red zero
		}
		$row .= '</td>'.$nl;

		## Presence in distributions

		my $total = scalar grep {$_} values %$distr_sel;
		my $present = 0;
		if ( $lib_obj->{DISTRS} ) {
			$present = scalar grep { $distr_sel->{$_} && $lib_obj->{DISTRS}{$_} } keys %$distr_sel;
		}
		my $approved = $lib_obj->{APPROVED};

		my $bgcolor;
		if ( $present == $total ) {
			if ( $filt_show && $filt_show eq 'problems' ) { next; } # drop the row
			$bgcolor = "#CCFFCC"; # green
		} elsif ( $present > 0 ) {
			$bgcolor = "#FEFECC"; # yellow
		} else {
			$bgcolor = "#FFCCCC"; # red
		}

		$row .= '<td align="right" style="background:'.$bgcolor.'">';
		my $link = href(tab_link('libs').'&list='.$soname);
		my $details_link = ' (<a href="'.html($link).'"'
			.' onclick="return popup_win(\''.html($link).'\')">list&hellip;</a>)';
		if ( $approved or $present == $total ) {
			if ( $filt_show && $filt_show eq 'uncommon' ) { next; } # drop the row
			$row .= $present." of ".$total;
			$row .= $details_link;
		} elsif ( $present > 0 ) {
			$row .= $present." known distro".($present > 1 ? 's' : '');
			$row .= $details_link;
		} else {
			$row .= 'Unknown library';
			# Tooltip:
			$row .= '&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, \'libs_ints_unknown\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
		}
		$row .= '</td>'.$nl;


		## LSB
		$row .= '<td align="center">';

		$row .= '<div>';

		my $lsb_status = $lib_obj->{LSB_STATUS}{STATUS};
		my $lsb_status_ver = $lib_obj->{LSB_STATUS}{STATUS_VER};
		
		my $color = undef;

		if ( $lsb_status eq "NOT" || $lsb_status eq "WITHDRAWN" ) {
			$row .= '<span style="color:red">No</span>';
		}
		elsif ( $lsb_status eq "CANDIDATE" ) {
			$row .= '<span style="color:#907000">Candidate for '.$lsb_status_ver.'</span>';
		}
		elsif ( $lsb_status eq "FUTURE" ) {
			$row .= '<span style="color:#907000">In LSB '.$lsb_status_ver.'</span>';
		}
		elsif ( $lsb_status eq "OPTIONAL" ) {
			if ( $filt_show && $filt_show eq 'non-lsb' ) { next; } # drop the row
			$row .= '<span style="color:blue">Optional</span>';
		}
		elsif ( $lsb_status eq "IN" ) {
			if ( $filt_show && $filt_show eq 'non-lsb' ) { next; } # drop the row
			$row .= '<span style="color:green">Yes</span>';
		}
		elsif ( $lsb_status eq "DEPRECATED" ) {
			if ( $filt_show && $filt_show eq 'non-lsb' ) { next; } # drop the row
			$row .= '<span style="color:blue">Deprecated</span>';
		}
		if ( $lsb_status ne "DEPRECATED" ) {
			if ( $filt_show && $filt_show eq 'deprecated' ) { next; }
		}

		$row .= '</div>';

		$row .= '</td>'.$nl;


		## More Info

		$row .= '<td>';
		my $txt = 'Details&hellip;';
		if ( $lib_obj->{MORE_INFO} && $lib_obj->{MORE_INFO} eq 'MORE_INFO' ) {
			$txt = '<b>More&nbsp;Info&hellip;</b>';
		}
		if ( $lib_obj->{MORE_INFO} && $lib_obj->{MORE_INFO} eq 'KNOWN_ISSUES' ) {
			$txt = '<b>Known&nbsp;Issues&hellip;</b>';
		}
		my $info_link = href( tab_link('libs').'&lib='.$soname );
		my $info_onclick = 'onclick="return popup_win(\''.html($info_link).'\')"';
		$row .= '<a href="'.html($info_link).'" '.$info_onclick.'>'.$txt.'</a>';
		$row .= '</td>'.$nl;

		$row .= '</tr>'.$nl;
		
		$table .= $row;
		$row_n++;
	}
	$table .= '</table>'.$nl;
	
	if ( $filtered && $row_n == 0 ) {
		$table .= '<div style="padding:10pt;text-align:center;font-size:large;color:#009">'
					.'<p>No libraries match the filter.</p>'
				.'</div>';
	}
	
	return ($table, $row_n);
}

#-----------------------------------------------------------------------

# Contents of a popup window with details on a library presence
# in a selected distribution subset.
sub lib_distr_list {
	my ($soname, $distr_sel, $filt_com) = @_;

	my $nl = "\n";
	my $res = "";

	my $libs = libs_data($distr_sel, $filt_com);
	my $lib_obj = $libs->{$soname};

	my $navi_link = href('http://www.linuxfoundation.org/navigator/browse/rawlib.php'
			.'?cmd=list-byrunname&RLrunname='.$soname);

	$res .= '<h1>Library: '
		.'<a href="'.html($navi_link).'" target="_blank">'.html($soname).'</a></h1>'.$nl;
	$res .= '<hr />'.$nl;
	
	# Presence in Distributions
	$res .= '<h2>Presence in Distributions</h2>'.$nl;

	$res .= lib_distr_list_sub($libs, $soname, $distr_sel);
	$res .= '<hr />'.$nl;
	
	return $res;
}

sub lib_distr_list_sub {
	my ($libs, $soname, $distr_sel) = @_;

	my $nl = "\n";
	my $res = "";

	my $lib_obj = $libs->{$soname};
	if ( !$lib_obj ) {
		$res .= "Error: there is no such library: '".html($soname)."'";
		return $res;
	}
	
	my $filtered = scalar grep {!$_} values %$distr_sel;
	my $approved = $lib_obj->{APPROVED};
	
	my $total_num = scalar grep {$_} values %$distr_sel;
	my $present_num = scalar grep {$_} values %{$lib_obj->{DISTRS}};
	my $missing_num = $total_num - $present_num;

	if ( $present_num > 0 ) {
		$res .= '<p>';
		$res .= 'Present in <span style="color:green">'.$present_num.'</span>';
		if ( $approved or $missing_num == 0) {
			$res .= ' and missing in <span style="color:red">'.$missing_num.'</span>';
		} else {
			$res .= ' and nothing is known about <span style="color:#907000">'.$missing_num.'</span>'; # yellow
		}
		if ( $filtered ) {
			$res .= ' of '.$total_num.' selected distributions.';
		} else {
			$res .= ' of total '.$total_num.' distributions.';
		}
		$res .= '</p>'.$nl;
	} else {
		if ( $approved ) {
			$res .= '<p>This library is missing in all the '.$total_num;
			if ( $filtered ) {
				$res .= ' selected distributions.</p>'.$nl;
			} else {
				$res .= ' analyzed distributions.</p>'.$nl;
			}
		} else {
			$res .= '<p>There is no information about this library in the database.</p>'.$nl;
		}
	}

	local *distr_list = sub {
		my ($color, @distrs) = @_;

		my $res = "";

		my $all_distrs = $Analysis::all_distrs;
		@distrs = sort {lc $all_distrs->{$a}{FULLNAME} cmp lc $all_distrs->{$b}{FULLNAME}} @distrs;
		$res .= '<ul>'.$nl;
		for my $Did ( @distrs ) {
			my $link = href(tab_link('distrs').'&id='.$Did);
			
			$res .= '<li><a style="color:'.$color.'" href="'.html($link).'">'
						.html($all_distrs->{$Did}{FULLNAME}).'</a></li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};

	if ( $present_num ) {
		$res .= '<h3>Present in:</h3>'.$nl;

		my @present_distrs = grep { $distr_sel->{$_} && $lib_obj->{DISTRS}{$_} } keys %$distr_sel;
		
		$res .= distr_list('green', @present_distrs);
	}

	if ( $missing_num ) {
		my @missing_distrs = grep { $distr_sel->{$_} && !$lib_obj->{DISTRS}{$_} } keys %$distr_sel;
		if ( $approved ) {
			$res .= '<h3>Missing in:</h3>'.$nl;
			
			$res .= distr_list('red', @missing_distrs);
		} else {
			$res .= '<h3>No information for:</h3>'.$nl;

			$res .= distr_list('#907000', @missing_distrs); # yellow
		}
	}

	return $res;
}

# Contents of a details popup window for Libraries table
sub lib_details {
	my ($soname, $selected_lsb_ver) = @_;

	my $nl = "\n";
	my $res = "";

	my $libs = libs_data();
	libs_lsb_data($libs, $selected_lsb_ver);
	my $lib_obj = $libs->{$soname};
	
	if ( !$lib_obj ) {
		$res .= "Error: there is no such library: '".html($soname)."'";
		return $res;
	}
	
	# Title
	my $navi_link = href( 'http://www.linuxfoundation.org/navigator/browse/rawlib.php'
			.'?cmd=list-byrunname&RLrunname='.$soname );
	$res .= '<h1>Library: <a href="'.html($navi_link).'" target="_blank">'.html($soname).'</a></h1>';

	$res .= '<hr/>'.$nl;

	# Check Lib Knowledge Base
	my $kb_info = lib_KB_info($soname);
	if ( $kb_info ) {
		$res .= '<h2>Known compatibility issues:</h2>'.$nl;
		$res .= '<p>'.$kb_info.'</p>'.$nl;

		$res .= '<hr />'.$nl;
	}

	# Check LibAttr
	my $attr = $Analysis::lib_attrs->{$soname};
	if ( $attr ) {
		foreach my $type ( sort keys %$attr ) {
			my $attr_html = "";
			my $bgcolor = '#fec';
			if ( $type eq "Recommendation" ) {
				$attr_html .= '<h3>LSB Recommendation</h3>';
				$attr_html .= html($attr->{$type});
			}
			elsif ( $type eq "StaticLinkage" ) {
				$attr_html .= '<h3>May be linked statically</h3>';
				$attr_html .= html($attr->{$type});
			}
			elsif ( $type eq "Rejection" ) {
				$attr_html .= '<h3>LSB Rejection</h3>';
				$attr_html .= html($attr->{$type});
			}
			else { # Custom
				$attr_html .= html($attr->{$type});
			}
			$attr_html =~ s/^\s+//; # remove unnecessary spaces
			$attr_html =~ s/\s+$//; #
			$attr_html .= "." if $attr_html =~ /[^\s\.]\s*$/; # Add a fullstop if necessary.
			
			$res .= '<div class="box" style="background:'.$bgcolor.'">'.$attr_html.'</div>';
		}

		$res .= '<hr />'.$nl;
	}

	# LSB Status
	$res .= '<h2>LSB Status:</h2>'.$nl;

	my ($lsb_status, $lsb_status_ver) = @{$lib_obj->{LSB_STATUS}}{qw/STATUS STATUS_VER/};

	if ( $lsb_status eq "NOT" ) {
		$res .= '<div style="color:red">Never been in LSB.</div>'.$nl;
	}
	elsif ( $lsb_status eq "FUTURE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. (Appeared in LSB '.$lsb_status_ver.')</div>'.$nl;
	}
	elsif ( $lsb_status eq "CANDIDATE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. Candidate for LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "OPTIONAL" ) {
		$res .= '<div style="color:blue">Optional since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "IN" ) {
		$res .= '<div style="color:green">Mandatory since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "DEPRECATED" ) {
		$res .= '<div style="color:blue">Deprecated since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "WITHDRAWN" ) {
		$res .= '<div style="color:red">Withdrawn since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}

	$res .= '<hr />'.$nl;

	# Presence in Distributions
	$res .= '<h2>Presence in Distributions</h2>'.$nl;
	
	$res .= lib_distr_list_sub($libs, $soname, $Analysis::all_distrs);

	$res .= '<hr />'.$nl;

	return $res;
}

# Returns information from the Knowledge Base for specified library and
#  distribution set.
sub lib_KB_info {
	my ($soname) = @_;

	my $nl = "\n";
	my $res = "";

	my $problems = $Analysis::libKBDistr->{$soname};

	return undef if not $problems;

	foreach my $LKBid ( sort {$a <=> $b} keys %$problems ) {
		my $text = $Analysis::libKB->[$LKBid];
		$text =~ s/\n/<br\/>\n/g;

		$res .= '<div>';
		$res .= "<p>";
		$res .= html($text);
		$res .= "</p>".$nl;

		my $grouped = $problems->{$LKBid};

		foreach my $i ( sort {$a <=> $b} keys %$grouped ) {
			$res .= '<div><b>Group '.$i.':</b> ';

			# Print IDs of distributions from this group
			$res .= join ", ", map {$Analysis::all_distrs->{$_}{FULLNAME}} @{$grouped->{$i}};

			$res .= '.</div>'.$nl;
		}

		$res .= '</div>'
	}

	return $res;
}

# Distribution compatibility table
sub D_tab {
	my ($selected_lsb_ver) = @_;
	
	my $res = "";
	my $nl = "\n";
	
	# Check if the table is empty
	if ( !%$Analysis::app_components ) {
		$res .= no_binary_div();
		return $res;
	}
	
	my $link = href(tab_link('distrs'));
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes'
		);
	w.moveTo(100,0);
	w.focus();

	return false;
}
// ]]>
</script>
HEREDOC

	$res .= <<DATA;
<div id="distro-details" style="position: absolute; border: 1px #aaaaaa solid; border-left: none; background: white; padding: 0px 2px; vertical-align: bottom; display: none;">
</div>

<p>The table below shows the compatibility status of your application with the distributions analyzed by the Linux Foundation. Your Application will run on the "<font color="green">green</font>" distributions without loader problems. Compatibility with the "<font color="#c0a000">yellow</font>" distributions can be easily achieved by excluding unneeded libraries from the dependencies of your application. Making the Application compatible with the "<font color="red">red</font>" distributions may require more effort to avoid using missing libraries/interfaces or by supplying them as a part of your application package. Please note that functional correctness is not guaranteed by this analysis.</p>
DATA

	my $ext_table_display = ( $_GET{'expand'} ? '' : 'none' );
	my $ext_table = <<DATA;
<table class="wikitable" cellpadding="0" id="d_table_ext" style="display: $ext_table_display; margin: 0; border-left: none;">
<tr><th colspan="2" style="border-left: none;">Missing&nbsp;Libraries</th><th colspan="2" style="">Missing&nbsp;Interfaces</th><th rowspan="2">Comments</th></tr>
<tr>
  <th style="width: 5em; border-left: none;"><table cellpadding="0" cellspacing="0" border="0" width="100%"><tr><td align="center" style="border: none; white-space: nowrap;">Common&nbsp;<a href="tests_help.pl#libs_ints_common" onclick="javascript:show_hide_tooltip(this, 'libs_ints_common', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></td></tr></table></th>
  <th style="width: 5em; white-space: nowrap;">Unknown&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, 'libs_ints_unknown', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
  <th style="width: 5em; white-space: nowrap;">Common&nbsp;<a href="tests_help.pl#libs_ints_common" onclick="javascript:show_hide_tooltip(this, 'libs_ints_common', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
  <th style="width: 5em; white-space: nowrap;">Unknown&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, 'libs_ints_unknown', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></th>
</tr>
DATA

	my $general_table = <<DATA;
<table class="wikitable" id="d_table" style="margin: 0;">
<tr><th colspan="2">Summary</th></tr>
<tr><th>Distribution</th><th style="width: 10em;"><table cellpadding="0" cellspacing="0" border="0"><tr><td align="left" style="border: none;"><img src="images/environment-spacer.gif" width="17" height="17" alt="" /></td><td width="100%" style="border: none;">Status</td><td align="right" style="border: none;"><img src="images/status_expand.png" width="16" height="16" alt="[&gt;&gt;]" id="expand-img" style="cursor: pointer;" onclick="javascript:toggle_table();" /></td></tr></table></th></tr>
DATA

	my $distrs = {};
	foreach my $Did ( keys %{$Analysis::all_distrs} ) {
		$distrs->{$Did} = Analysis::distr_data($Did);
	}
	
	local *by_status = sub {
		my ($a, $b) = @_;
		
		return (
			$Analysis::status_range{$a->{STATUS}} <=> $Analysis::status_range{$b->{STATUS}}
			or $a->{MISSING_LIBS}   <=> $b->{MISSING_LIBS}
			or $a->{MISSING_INTS}   <=> $b->{MISSING_INTS}
			or $a->{UNNEEDED_LIBS} <=> $b->{UNNEEDED_LIBS}
			or $a->{UNKNOWN_LIBS}  <=> $b->{UNKNOWN_LIBS}
			or $a->{UNKNOWN_INTS}  <=> $b->{UNKNOWN_INTS}
			or lc $a->{FULLNAME} cmp lc $b->{FULLNAME}
		);
	};
	
	# Distributions list
	foreach my $distr ( sort {by_status($a,$b)} values %$distrs ) {
		my $Did = $distr->{ID};
		my $status = $distr->{STATUS};

		my $bgcolor;
		my $classname;
		if ( $status eq 'MISSING' ) {
			$bgcolor = "#ffdddd"; # red
			$classname = 'fail-td';
			$status = 'INCOMPATIBLE';
		}
		elsif ( $status eq 'UNKNOWN' || $status eq 'UNNEEDED' ) {
			$bgcolor = "#fefecc"; # yellow
			$classname = 'warn-td';
			$status = 'WARNINGS';
		}
		else {
			$bgcolor = "#ccffcc"; # green
			$classname = 'pass-td';
			$status = 'OK';
		}
		
		my $missing_libs   = $distr->{MISSING_LIBS};
		my $unneeded_libs = $distr->{UNNEEDED_LIBS};
		my $unknown_libs  = $distr->{UNKNOWN_LIBS};
		
		my $missing_ints  = $distr->{MISSING_INTS};
		my $unknown_ints = $distr->{UNKNOWN_INTS};
		
		# Distribution
		my $Dname = html($distr->{FULLNAME});
		my $Dname_link = html(href($distr->{NAME}));
		$general_table .= <<DATA;
<tr id="dtr_$Did" style="background: $bgcolor;">
<td style="white-space: nowrap;"><b><a href="http://www.linuxfoundation.org/navigator/browse/distr.php?cmd=list-byname&amp;Dname=$Dname_link" target="_blank">$Dname</a></b></td>
<td align="center" class="$classname" onmouseover="javascript:show_info(this, $missing_libs, $unneeded_libs, $missing_ints, $unknown_libs, $unknown_ints);" onmouseout="javascript:hide_info();" onclick="javascript:toggle_table();"><b>$status</b></td>
</tr>
DATA

		$ext_table .= '<tr style="background: '.$bgcolor.';">';

		# Missing libraries
		if ($missing_libs + $unneeded_libs == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			my $unneeded_libs_html = ($unneeded_libs ? ' (+<font color="red"><b>'.$unneeded_libs.'</b></font> unneeded)' : '');
			my $missing_libs_html = ($missing_libs ? '<font color="red"><b>'.$missing_libs.'</b></font>' : '0');
			my $link = href(tab_link('distrs').'&id='.$Did.'&show=missing_libs');
			$ext_table .= '<td style="white-space: nowrap; border-left: none;">'
				.$missing_libs_html.$unneeded_libs_html
				.' (<a href="'.html($link).'" onclick="return popup_win(\''.html($link).'\')"'
					.'>list&hellip;</a>)'
				.'</td>';
		}

		# Unknown libraries
		if ($unknown_libs == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			my $link = href(tab_link('distrs').'&id='.$Did.'&show=unknown_libs');
			$ext_table .= '<td style="white-space: nowrap; border-left: none;">'
				.'<font color="red"><b>'.$unknown_libs.'</b></font>'
				.' (<a href="'.html($link).'" onclick="return popup_win(\''.html($link).'\')"'
					.'>list&hellip;</a>)'
				.'</td>';
		}

		# Missing interfaces
		if ($missing_ints == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			my $link = href(tab_link('distrs').'&id='.$Did.'&show=missing_ints');
			$ext_table .= '<td style="white-space: nowrap; border-left: none;">'
				.'<font color="red"><b>'.$missing_ints.'</b></font>'
				.' (<a href="'.html($link).'" onclick="return popup_win(\''.html($link).'\')"'
					.'>list&hellip;</a>)'
				.'</td>';
		}

		# Unknown interfaces
		if ($unknown_ints == 0) {
			$ext_table .= '<td style="border-left: none; text-align: center; color: gray;">none</td>';
		}
		else {
			my $link = href(tab_link('distrs').'&id='.$Did.'&show=unknown_ints');
			$ext_table .= '<td style="white-space: nowrap; border-left: none;">'
				.'<font color="red"><b>'.$unknown_ints.'</b></font>'
				.' (<a href="'.html($link).'" onclick="return popup_win(\''.html($link).'\')"'
					.'>list&hellip;</a>)'
				.'</td>';
		}
		
		# Comments
		my $comments = '';
		if ( $distr->{COMMENTS} && $distr->{COMMENTS}{NO_GNU_HASH} ) {
			$comments = '<span style="white-space:nowrap">No .gnu.hash support <a href="" onclick="javascript:show_hide_tooltip(this, \'com_no_hash_section\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></span>';
		}
		$ext_table .= '<td>'.$comments.'</td>'.$nl;
		
		$ext_table .= '</tr>'.$nl;
	}
	$general_table .= '</table>';
	$ext_table .= '</table>';

	$res .= <<HEREDOC;
<table cellpadding="0" cellspacing="0" border="0" style="margin: 12px 0;">
<tr>
<td>
$general_table
</td>
<td>
$ext_table
</td>
</tr>
</table>

<script type="text/javascript" language="javascript">
// <![CDATA[

// Determines absolute dynamic position of an element (right top corner).
function getPosition(elem) {
	if (!elem)
		return {x: -1, y: -1};
	var x = elem.offsetWidth;
	var y = 0;
	while (elem) {
		x += elem.offsetLeft;
		y += elem.offsetTop;
		elem = elem.offsetParent;
	}
	return {x: x, y: y};
}

var distro_details = document.getElementById('distro-details');
var table = document.getElementById('d_table');
var table_ext = document.getElementById('d_table_ext');
var expand_img = document.getElementById('expand-img');

// Fine-tuning for browsers
var dDetailsOffsetX = ((navigator.appName == 'Netscape') ? 0 : 1);
if (navigator.appName == 'Konqueror') {
	table_ext.style.borderLeft = 'solid 1px #aaaaaa';
}

function show_info(td_elem, missing_libs, missing_zero_libs, missing_ints, unknown_libs, unknown_ints) {
	if (table_ext.style.display != 'none')
		return;
	var details = '';

	if (missing_libs + missing_zero_libs + missing_ints > 0) {
		var tmp = ''
		if (missing_libs + missing_zero_libs > 0) {
			tmp += (missing_libs ? ('<font color="red"><b>' + missing_libs + '</b></font> ' + ((missing_libs > 1) ? 'libraries' : 'library')) : '0 libraries') +
			       (missing_zero_libs ? ' (+<font color="red"><b>' + missing_zero_libs + '</b></font> unneeded)' : '');
		}
		if (missing_ints > 0) {
			tmp += (tmp == '' ? '' : '; ') +
			       (missing_ints ? ('<font color="red"><b>' + missing_ints + '</b></font> ' + ((missing_ints > 1) ? 'interfaces' : 'interface')) : '0 interfaces');
		}
		details += 'Missing: ' + tmp + '.';
	}
	if (unknown_libs + unknown_ints > 0) {
		var tmp = ''
		if (unknown_libs > 0) {
			tmp += (unknown_libs ? ('<font color="red"><b>' + unknown_libs + '</b></font> ' + ((unknown_libs > 1) ? 'libraries' : 'library')) : '0 libraries');
		}
		if (unknown_ints > 0) {
			tmp += (tmp == '' ? '' : '; ') +
			       (unknown_ints ? ('<font color="red"><b>' + unknown_ints + '</b></font> ' + ((unknown_ints > 1) ? 'interfaces' : 'interface')) : '0 interfaces');
		}
		details += (details == '' ? '' : ' ') + 'Unknown: ' + tmp + '.';
	}
	if (details == '')
		details = '<i>No problems detected.</i>';
	distro_details.innerHTML = details;
	var pos = getPosition(td_elem);
	distro_details.style.left = (pos.x + dDetailsOffsetX) + 'px';
	distro_details.style.top = pos.y + 'px';
	distro_details.style.height = (td_elem.offsetHeight - 1) + 'px';
	distro_details.style.display = '';
}

function hide_info() {
	distro_details.style.display = 'none';
}

function toggle_table() {
	hide_info();
	if (table_ext.style.display == 'none') {
		expand_img.setAttribute('src', 'images/status_collapse.png');
		expand_img.setAttribute('alt', '[&lt;&lt;]');
		table_ext.style.display = '';
	}
	else {
		expand_img.setAttribute('src', 'images/status_expand.png');
		expand_img.setAttribute('alt', '[&gt;&gt;]');
		table_ext.style.display = 'none';
	}
}
// ]]>
</script>
HEREDOC

	return $res;
} # end of  D_tab
#-----------------------------------------------------------------------

# Contents of a detail popup window for Distributions table
sub distr_details {
	my ($Did, $show) = @_;

	my $nl = "\n";
	my $res = "";

	my $distr = Analysis::distr_data($Did);
	if ( !$distr ) {
		$res .= "There is no distribution with ID '$Did'.";
		return $res;
	}
	
	# The header
	my $navi_link = href('http://www.linuxfoundation.org/navigator/browse/distr.php'
			.'?cmd=list-byname&Dname='.$distr->{NAME} );
	$res .= '<h1>Distribution: '
		.'<a href="'.html($navi_link).'" target="_blank">'.html($distr->{FULLNAME}).'</a>'
		.'</h1>'.$nl;
	$res .= '<hr/>';
	
	if ( !$show || $show eq 'other' ) {
		if ( $distr->{NO_GNU_HASH} ) {
			if ( grep { $_->{NO_SHT_HASH} } values %$Analysis::app_components ) {
				$res .= '<h2>Section .gnu.hash is not supported</h2>'.$nl;
				$res .= '<p>Binaries targeting this distribution must use the \'.hash\' section instead.</p>'.$nl;
				$res .= '<hr/>';
			}
		}
	}
	
	# App components
	if ( %$Analysis::app_components && (!$show || $show eq 'components') ) {
		$res .= '<h2>Application components:</h2>'.$nl;
		$res .= '<p>The table below shows the compatibility information for application components with regard to '.html($distr->{FULLNAME}).'.</p>'.$nl;
		$res .= '<table style="border-collapse:collapse" border="1">'.$nl;
		$res .= '<tr>'
				.'<th rowspan="1">Application component</th>'
				.'<th colspan="2">Libraries</th>'
				.'<th colspan="2">Interfaces</th>'
			.'</tr>'.$nl;
		$res .= '<tr>'
				.'<th></th>'
				.'<th>Missing</th>'
				.'<th>Unknown</th>'
				.'<th>Missing</th>'
				.'<th>Unknown</th>'
			.'</tr>'.$nl;
		foreach my $com ( sort keys %$Analysis::app_components ) {
			my ($disp_name, $disp_path) = parse_app_component($com);
			my $com_obj = $Analysis::app_components->{$com};
			Analysis::com_distr_data($com_obj, $Did, $com_obj->{LIBS_UNNEEDED});
			my $status = Analysis::com_distr_status($com_obj, $Did);
			my $bgcolor = "";
			if ( $status eq 'OK' ) {
				$bgcolor = "#ccffcc"; # green
			} elsif ( $status eq 'UNNEEDED' ) {
				$bgcolor = "#fefecc"; # yellow
			} elsif ( $status eq 'UNKNOWN' ) {
				$bgcolor = "#fefecc"; # yellow
			} elsif ( $status eq 'MISSING' ) {
				$bgcolor = "#ffdddd"; # red
			}
			my $link = href(tab_link('com').'&com='.$com.'&distr='.$Did);
			
			# Component Info
			my ($missing_libs, $unneeded_libs, $unknown_libs, $missing_ints, $unknown_ints) = (0) x 5;
			foreach my $req_soname ( @{$com_obj->{NEEDED_LIBS}} ) {
				my $val = $com_obj->{$Did}{LIBS}{$req_soname}{STATUS};
				if ( $val eq 'MISSING' ) { $missing_libs++; }
				if ( $val eq 'UNNEEDED' ) { $unneeded_libs++; }
				if ( $val eq 'UNKNOWN' ) { $unknown_libs++; }
			}
			foreach my $val ( values %{$com_obj->{$Did}{INT_STATUS}} ) {
				if ( $val eq 'MISSING' ) { $missing_ints++ }
				if ( $val eq 'UNKNOWN' ) { $unknown_ints++; }
			}
			
			$res .= '<tr style="background:'.html($bgcolor).'">';
			$res .= '<td><a href="'.html($link).'" title="'.html($disp_path).'">'.html($disp_name).'</a></td>';
			
			$res .= '<td>';
			if ( $missing_libs ) {
				$res .= '<span style="color:red">';
				$res .= $missing_libs.($unneeded_libs ? ' (+'.$unneeded_libs.' unneeded)' : '');
				$res .= '</span>';
			} elsif ( $unneeded_libs ) {
				$res .= '<span style="color:#907000">';
				$res .= $unneeded_libs.' unneeded';
				$res .= '</span>';
			}
			$res .= '</td>';
			$res .= '<td><span style="color:#907000">'.($unknown_libs or "").'</span></td>';
			$res .= '<td><span style="color:red">'.($missing_ints or "").'</span></td>';
			$res .= '<td><span style="color:#907000">'.($unknown_ints or "").'</span></td>';
			$res .= '</tr>'.$nl;
		}
		$res .= '</table>'.$nl;
		$res .= '<hr/>'.$nl;
	}

	# Libraries
	local *lib_list = sub {
		my ($color, @sonames) = @_;

		my $res = "";

		$res .= '<ul>'.$nl;
		foreach my $soname ( sort {lc $a cmp lc $b} @sonames ) {
			my $link = href(tab_link('libs').'&lib='.$soname);
			$res .= '<li>';
			# Not a link, because the link won't be useful here
			$res .= '<span style="color:'.html($color).'">'.html($soname).'</span>';
			$res .= '</li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};
	
	my $hr_after_libs = 0;
	
	if ( $distr->{MISSING_LIBS} && (!$show || $show eq 'missing_libs') ) {
		$res .= '<h2>Missing Libraries:</h2>'.$nl;

		$res .= lib_list( 'red', grep {$distr->{LIBS}{$_} eq 'MISSING'} keys %{$distr->{LIBS}} );
		$hr_after_libs = 1;
	}

	if ( $distr->{UNKNOWN_LIBS} && (!$show || $show eq 'unknown_libs') ) {
		$res .= '<h2>No Information About:</h2>'.$nl;
		$res .= '<p>The Application Checker Database contains no information'
			.' about the following libraries:</p>'.$nl;
		# yellow
		$res .= lib_list('#907000', grep {$distr->{LIBS}{$_} eq 'UNKNOWN'} keys %{$distr->{LIBS}});
		$hr_after_libs = 1;
	}
	
	if ( $distr->{UNNEEDED_LIBS} && (!$show || $show eq 'unneeded_libs' || $show eq 'missing_libs') ) {
		$res .= '<h2>Unnecessary Libraries:</h2>'.$nl;
		$res .= '<p>These are libraries required by the Application,'
			.' but <b>no interfaces</b> from these libraries are actually used:</p>'.$nl;
		# yellow
		$res .= lib_list('#907000', grep {$distr->{LIBS}{$_} eq 'UNNEEDED'} keys %{$distr->{LIBS}});
		$hr_after_libs = 1;
	}
	
	if ( $distr->{OK_LIBS} && (!$show || $show eq 'present_libs') ) {
		$res .= '<h2>Present Libraries:</h2>'.$nl;

		$res .= lib_list('green', grep {$distr->{LIBS}{$_} eq 'OK'} keys %{$distr->{LIBS}});
		$hr_after_libs = 1;
	}

	$res .= '<hr/>' if $hr_after_libs;
	
	local *int_list = sub {
		my ($color, @fullnames) = @_;

		my $res = "";

		$res .= '<ul>'.$nl;
		foreach my $fullname ( sort {lc $a cmp lc $b} @fullnames ) {
			$res .= '<li>';
			# Not a link, because the link won't be useful here
			$res .= '<span style="color:'.html($color).'">'.html($fullname).'</span>';
			$res .= '</li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};

	my $hr_after_ints = 0;
	
	if ( $distr->{MISSING_INTS} && (!$show || $show eq 'missing_ints') ) {
		$res .= '<h2>Missing Interfaces:</h2>'.$nl;

		$res .= int_list('red', grep {$distr->{INTS}{$_} eq 'MISSING'} keys %{$distr->{INTS}});
		$hr_after_ints = 1;
	}
	
	if ( $distr->{UNKNOWN_INTS} && (!$show || $show eq 'unknown_ints') ) {
		$res .= '<h2>No Information About:</h2>'.$nl;
		$res .= '<p>The Application Checker Database contains no information'
			.' about the following interfaces:</p>'.$nl;
		# yellow
		$res .= int_list('#907000', grep {$distr->{INTS}{$_} eq 'UNKNOWN'} keys %{$distr->{INTS}});
		$hr_after_ints = 1;
	}
	# OK_INTS are not shown
	
	$res .= '<hr/>' if $hr_after_ints;
	
	if ( $show ) { # if filtered...
		my $link = href(tab_link('distrs').'&id='.$Did);
		$res .= '<p><a href="'.html($link).'"><i>More details&hellip;</i></a></p>';
	}
	
	return $res;
}

sub summary_data {
	my ($selected_lsb_ver) = @_;
	
	my $summary = {};
	
	$summary->{STATUS} = Analysis::overall_verdict();
	
	# Calculate statistics on distribution statuses
	$summary->{DISTRS}{'TOTAL'} = scalar keys %$Analysis::all_distrs;
	@{$summary->{DISTRS}}{'GOOD', 'WARN', 'BAD'} = (0)x3; # initialize the statistic
	foreach my $Did ( keys %$Analysis::all_distrs ) {
		my $status = Analysis::distr_status($Did); # OK|UNNEEDED|UNKNOWN|MISSING
		if ( $status eq 'OK' ) { $summary->{DISTRS}{GOOD}++; }
		elsif ( $status eq 'UNNEEDED' || $status eq 'UNKNOWN' ) { $summary->{DISTRS}{WARN}++; }
		elsif ( $status eq 'MISSING' ) { $summary->{DISTRS}{BAD}++; }
	}
	
	# Select the displayed status of distributions
	my $dstat = $summary->{DISTRS};
	if ( $dstat->{WARN} == 0 && $dstat->{BAD} == 0 ) {
		$dstat->{STATUS} = 'ALL';
	}
	elsif ( $dstat->{WARN} > 0 && $dstat->{BAD} == 0 ) {
		$dstat->{STATUS} = 'ALL_EXCEPT';
	}
	elsif ( $dstat->{GOOD} > 0 && $dstat->{WARN} == 0 && $dstat->{BAD} > 0 ) {
		$dstat->{STATUS} = 'SOME_GOOD';
	}
	elsif ( $dstat->{WARN} > 0 && $dstat->{BAD} > 0 ) {
		$dstat->{STATUS} = 'SOME_GOOD_EXCEPT';
	}
	elsif ( $dstat->{GOOD} == 0 && $dstat->{WARN} == 0 && $dstat->{BAD} > 0 ) {
		$dstat->{STATUS} = 'NO_ONE';
	}
	
	# Check the hash type
	foreach my $com_obj ( values %$Analysis::app_components ) {
		if ( $com_obj->{NO_SHT_HASH} ) {
			$summary->{NO_SHT_HASH}++;
		}
	}

	# Libraries
	my $libs = libs_data();
	libs_lsb_data($libs, $selected_lsb_ver);

	# initialization
	$summary->{ZERO_INT_LIBS} = 0;
	$summary->{UNAPPROVED_LIBS} = 0;
	for ( qw/NOT CANDIDATE FUTURE IN OPTIONAL DEPRECATED WITHDRAWN/ ) {
		$summary->{LSB_LIBS}{$_}{COUNT} = 0;
		$summary->{LSB_LIBS}{$_}{VERS} = {};
	}
	
	foreach my $lib ( values %$libs ) {
		# Check libs without interfaces used
		if ( !$lib->{INTERFACES} || !%{$lib->{INTERFACES}} ) {
			$summary->{ZERO_INT_LIBS}++;
		}
		if ( !$lib->{APPROVED} ) {
			$summary->{UNAPPROVED_LIBS}++;
		}

		# Calculate LSB statistics for libraries
		my ($lsb_status, $lsb_status_ver) = @{$lib->{LSB_STATUS}}{'STATUS', 'STATUS_VER'};
		$summary->{LSB_LIBS}{$lsb_status}{COUNT}++;
		$summary->{LSB_LIBS}{$lsb_status}{VERS}{$lsb_status_ver}++ if $lsb_status_ver;
	}
	
	# Interfaces
	my $ints = ints_data();
	ints_lsb_data($ints, $selected_lsb_ver);
	
	for ( qw/NOT CANDIDATE FUTURE IN OPTIONAL DEPRECATED WITHDRAWN/ ) {
		$summary->{LSB_INTS}{$_}{COUNT} = 0;
		$summary->{LSB_INTS}{$_}{VERS} = {};
	}
	foreach my $Iname ( keys %$ints ) {
	foreach my $ver ( keys %{$ints->{$Iname}} ) {
	foreach my $lib_soname ( keys %{$ints->{$Iname}{$ver}{LIBS}} ) {
		my $int = $ints->{$Iname}{$ver}{LIBS}{$lib_soname};
		
		# Calculate LSB statistics for interfaces
		my ($lsb_status, $lsb_status_ver) = @{$int->{LSB_STATUS}}{'STATUS', 'STATUS_VER'};
		$summary->{LSB_INTS}{$lsb_status}{COUNT}++;
		$summary->{LSB_INTS}{$lsb_status}{VERS}{$lsb_status_ver}++ if $lsb_status_ver;
	}}}
	
	## Calculate the overall LSB status
	
	my $lib_lsb_status = $summary->{LSB_LIBS};
	my $int_lsb_status = $summary->{LSB_INTS};
	
	if ( $lib_lsb_status->{"NOT"}{COUNT} || $lib_lsb_status->{"WITHDRAWN"}{COUNT} ) {
		$summary->{LSB} = 'NON_LSB_LIBS';
	}
	elsif ( $int_lsb_status->{"NOT"}{COUNT} || $int_lsb_status->{"WITHDRAWN"}{COUNT} ) {
		$summary->{LSB} = 'NON_LSB_INTS';
	}
	elsif ( $lib_lsb_status->{"CANDIDATE"}{COUNT} ) {
		if ( $lib_lsb_status->{"FUTURE"}{COUNT} ) {
			$summary->{LSB} = 'CANDIDATE_FUTURE_LIBS';
		} else {
			$summary->{LSB} = 'CANDIDATE_LIBS';
		}
	}
	elsif ( $int_lsb_status->{"CANDIDATE"}{COUNT} ) {
		if ( $int_lsb_status->{"FUTURE"}{COUNT} ) {
			$summary->{LSB} = 'CANDIDATE_FUTURE_INTS';
		} else {
			$summary->{LSB} = 'CANDIDATE_INTS';
		}
	}
	elsif ( $lib_lsb_status->{"FUTURE"}{COUNT} ) {
		$summary->{LSB} = 'FUTURE_LIBS';
	}
	elsif ( $int_lsb_status->{"FUTURE"}{COUNT} ) {
		$summary->{LSB} = 'FUTURE_INTS';
	}
	elsif ( $lib_lsb_status->{"DEPRECATED"}{COUNT} || $int_lsb_status->{"DEPRECATED"}{COUNT} ) {
		$summary->{LSB} = 'DEPRECATED';
	}
	else { #  "IN" or "OPTIONAL"
		$summary->{LSB} = 'ALL_OK';
	}
	
	if ( !%$libs && !%$ints ) {
		$summary->{DISTRS}{STATUS} = 'NO_BINS';
		$summary->{LSB} = 'NO_BINS';
	}
	
	if ( $Analysis::improper_arch ) {
		$summary->{IMPROPER_ARCH} = $Analysis::improper_arch;
	}
	
	return $summary;
}
#-----------------------------------------------------------------------

# Renders a short summary for the header of the results page.
sub Summary {
	my ($selected_lsb_ver) = @_;
	
	my $summary = summary_data($selected_lsb_ver);
	
	## Let's print the summary

	my $nl = "\n";
	my $res = '';

	my $green_image  = '<img src="images/dot_green.png" width="16" height="16" alt="green" /> ';
	my $yellow_image = '<img src="images/dot_yellow.png" width="16" height="16" alt="yellow" /> ';
	my $red_image    = '<img src="images/dot_red.png" width="16" height="16" alt="red" /> ';
	my $warn_image   = '<img src="images/dot_warn.png" width="16" height="16" alt="warning" /> ';

	my $green_table = <<DATA;
<div style="background-color: #40d040; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>No compatibility problems detected</b>
</div>
<div id="summary" class="summary" style="background-color: #e6ffe6;">
DATA
	my $yellow_table = <<DATA;
<div style="background-color: #d0a720; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>Some compatibility problems detected</b>
</div>
<div id="summary" class="summary" style="background-color: #fff9d4;">
DATA
	my $red_table = <<DATA;
<div style="background-color: #d04040; border: 1px solid #7f7f7f; padding: 0.3em; text-align: left; vertical-align: middle; font-size: large; color: #ffffff; margin-bottom: 0;"><b>Numerous compatibility problems found</b>
</div>
<div id="summary" class="summary" style="background-color: #ffeeeb;">
DATA

	local *_pl = sub {
		my ($num, $mid, $singular, $plural) = @_;
		
		return '<b>'.$num.'</b> '.($mid ? $mid.' ' : '').($num > 1 ? $plural : $singular);
	};
	
	local *refer = sub {
		my ($text, $tab, $add, $no_below) = @_;
		
			my $link = href(tab_link($tab).'&nofilt=1');
			if ( $add ) { $link .= href($add); }
			return '<a href="'.html($link).'">'.$text.'</a>';
	};
	
	local *see_details = sub {
		my ($tab, $add) = @_;
		
		my $link = href(tab_link($tab));
		if ( $add ) { $link .= href($add); }
		return ' (<a href="'.html($link).'">details&hellip;</a>)';
	};

	# Summary block
	if ( $summary->{STATUS} eq 'Success' ) {
		$res .= $green_table;
	}
	elsif ( $summary->{STATUS} eq 'Warnings' ) {
		$res .= $yellow_table;
	}
	elsif ( $summary->{STATUS} eq 'Failed' ) {
		$res .= $red_table;
	}

	# Summary on compatibility with distributions
	my $dstat = $summary->{DISTRS};
	my $distr_status = $dstat->{STATUS};
	if ( $distr_status eq 'ALL' ) {
		$res .= '<p>';
		$res .= $green_image; # Green image
		$res .= 'All libraries and interfaces used by the Application';
		$res .= ' are present in all the analyzed distributions';
		$res .= see_details('distrs', '&expand=1');
		$res .= '.</p>'.$nl;
	}
	elsif ( $distr_status eq 'ALL_EXCEPT' ) {
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		$res .= 'All libraries and interfaces required by the Application'
				.' (except some unused or unknown libraries)'
				.' are present in all the analyzed distributions';
		$res .= see_details('distrs', '&expand=1');
		$res .= '.</p>'.$nl;
	}
	elsif ( $distr_status eq 'SOME_GOOD' ) {
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		$res .= 'There '.($dstat->{GOOD} > 1 ? 'are' : 'is' ).' '
				.refer('<b>'.$dstat->{GOOD}.'</b> of <b>'.$dstat->{TOTAL}.'</b> distributions', 'distrs', '&expand=1')
				.' that provide all the required libraries and interfaces.';
		$res .= '</p>'.$nl;
	}
	elsif ( $distr_status eq 'SOME_GOOD_EXCEPT' ) {
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		my $n = $dstat->{GOOD} + $dstat->{WARN};
		$res .= 'There '.($n > 1 ? 'are' : 'is' ).' '
				.refer('<b>'.$n.'</b> of <b>'.$dstat->{TOTAL}.'</b> distributions', 'distrs', '&expand=1')
				.' that provide all the required libraries and interfaces'
				.' except unneeded and unknown.';
		$res .= '</p>'.$nl;
	}
	elsif ( $distr_status eq 'NO_ONE' ) {
		$res .= '<p>';
		$res .= $red_image; # Red image
		$res .= 'The Application is unlikely to run on any of the analyzed distributions.';
		$res .= '</p>'.$nl;
	}
	elsif ( $distr_status eq 'NO_BINS' ) {
		if ( $summary->{IMPROPER_ARCH} ) {
			$res .= '<p>';
			$res .= $yellow_image; # Yellow image
			$res .= 'The Application Checker is unable to test applications for <b>another architecture</b>.';
			$res .= ' Please use the version of the Application Checker packaged for the same architecture as the program you want to check.';
			$res .= '</p>'.$nl;
		} else {
			$res .= '<p>';
			$res .= $yellow_image; # Green image
			$res .= 'The application requires no external binaries or interfaces.';
			$res .= ' Please see the LSB Certification tab for the results of script analysis (Perl, Python, Shell scripts).';
			$res .= '</p>'.$nl;
		}
	}
	
	if ( $summary->{IMPROPER_ARCH} && $distr_status ne 'NO_BINS' ) {
		$res .= '<p>';
		$res .= $yellow_image; # Yellow image
		$res .= 'The application includes binary components compiled for <b>another architecture</b>. These binaries will be ignored.';
		$res .= ' Please use the version of the Application Checker packaged for the same architecture as the program you want to check.';
		$res .= '</p>'.$nl;
	}

	# Summary on LSB compatibility
	$res .= '<p>';
	my $lib_lsb_status = $summary->{LSB_LIBS};
	my $int_lsb_status = $summary->{LSB_INTS};
	if ( $summary->{LSB} eq 'NON_LSB_LIBS' ) {
		my $x = $lib_lsb_status->{"NOT"}{COUNT} + $lib_lsb_status->{"WITHDRAWN"}{COUNT}
				+$lib_lsb_status->{"CANDIDATE"}{COUNT} + $lib_lsb_status->{"FUTURE"}{COUNT};
		$res .= $red_image; # Red image
		$res .= 'The Application uses '
			.refer( _pl($x, 'external', 'library', 'libraries'), 'libs', '&filt_show=non-lsb' )
			.' <b>incompatible with LSB '.$selected_lsb_ver.'</b>.';
	}
	elsif ( $summary->{LSB} eq 'NON_LSB_INTS' ) {
		my $x = $int_lsb_status->{"NOT"}{COUNT} + $int_lsb_status->{"WITHDRAWN"}{COUNT}
				+$int_lsb_status->{"CANDIDATE"}{COUNT} + $int_lsb_status->{"FUTURE"}{COUNT};
		$res .= $red_image; # Red image
		$res .= 'The Application uses '
			.refer( _pl($x, '', 'interface', 'interfaces'), 'ints', '&filt_show=non-lsb')
			.' <b>incompatible with LSB '.$selected_lsb_ver.'</b>.';
	}
	elsif ( $summary->{LSB} eq 'CANDIDATE_FUTURE_LIBS' ) {
		my $cand_version = (sort keys %{$lib_lsb_status->{"CANDIDATE"}{VERS}})[-1]; # max
		$res .= $yellow_image; # Yellow image
		$res .= 'All libraries used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or future versions or are candidates'
				.' for LSB '.$cand_version.'.';
		if ( $lib_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $summary->{LSB} eq 'CANDIDATE_LIBS' ) {
		my $cand_version = (sort keys %{$lib_lsb_status->{"CANDIDATE"}{VERS}})[-1]; # max
		$res .= $yellow_image; # Yellow image
		$res .= 'All libraries used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or are candidates'
				.' for LSB '.$cand_version.'.';		
		if ( $lib_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $summary->{LSB} eq 'CANDIDATE_FUTURE_INTS' ) {
		my $cand_version = (sort keys %{$int_lsb_status->{"CANDIDATE"}{VERS}})[-1]; # max
		$res .= $yellow_image; # Yellow image
		$res .= 'All interfaces used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or future versions or are candidates'
				.' for LSB '.$cand_version.'.';
		if ( $lib_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		} elsif ( $int_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some interfaces used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $summary->{LSB} eq 'CANDIDATE_INTS' ) {
		my $cand_version = (sort keys %{$int_lsb_status->{"CANDIDATE"}{VERS}})[-1]; # max
		$res .= $yellow_image; # Yellow image
		$res .= 'All interfaces used by the Application are either included'
				.' in LSB '.$selected_lsb_ver.' or are candidates'
				.' for LSB '.$cand_version.'.';	
		if ( $lib_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>.';
		} elsif ( $int_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some interfaces used by the Application are <b>deprecated</b>.';
		}
	}
	elsif ( $summary->{LSB} eq 'FUTURE_LIBS' ) {
		my $future_version = (sort keys %{$lib_lsb_status->{"FUTURE"}{VERS}})[-1]; # max
		
		my $x = $lib_lsb_status->{"FUTURE"}{COUNT};
		
		$res .= $yellow_image; # Yellow image
		$res .= 'The Application uses '._pl($x, '', 'library', 'libraries')
				.', which appeared in a <b>later version of the LSB</b>.';
		$res .= ' Try checking the Application for compatibility with <b>LSB '.$future_version.'</b>.';
		
		if ( $lib_lsb_status->{'DEPRECATED'} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
	}
	elsif ( $summary->{LSB} eq 'FUTURE_INTS' ) {
		my $future_version = (sort keys %{$int_lsb_status->{"FUTURE"}{VERS}})[-1]; # max
		
		my $x = $int_lsb_status->{"FUTURE"}{COUNT};
		
		$res .= $yellow_image; # Yellow image
		$res .= 'The Application uses '._pl($x, '', 'interface', 'interfaces')
				.', which appeared in <b>later version of the LSB</b>.';
		$res .= ' Try checking the Application for compatibility with <b>LSB '.$future_version.'</b>.';
		
		if ( $lib_lsb_status->{"DEPRECATED"} ) {
			$res .= ' Note that some libraries used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
		elsif ( $int_lsb_status->{"DEPRECATED"} ) {
			$res .= ' Note that some interfaces used by the Application are <b>deprecated</b>'
					.' in LSB '.$selected_lsb_ver.'.';
		}
	}
	elsif ( $summary->{LSB} eq 'DEPRECATED' ) {
		$res .= $yellow_image; # Yellow image
		$res .= 'All libraries and interfaces used by the Application';
		$res .= ' are <b>included in LSB '.$selected_lsb_ver.'</b>';
		$res .= ', but some of them are <b>deprecated</b>.';
		$res .= ' You may have compatibility issues with new distributions in the future.';		
	}
	elsif ( $summary->{LSB} eq 'ALL_OK' ) {
		$res .= $green_image; # Green image
		$res .= 'All libraries and interfaces used by the Application';
		$res .= ' are <b>included in LSB '.$selected_lsb_ver.'</b>.';
		$res .= ' You have a good chance to achieve '.refer('LSB compliance', 'cert', '', 1).'.';
	}
	elsif ( $summary->{LSB} eq 'NO_BINS' ) {
		# do nothing
	}
	else {
		die "Unhandled LSB summary: ".$summary->{LSB};
	}
	$res .= '</p>'.$nl;

	# Libraries without interfaces used
	if ( $summary->{ZERO_INT_LIBS} ) {
		$res .= '<p>';
		$res .= $warn_image; # Yellow image
		$res .= 'The Application uses '
			.refer( _pl($summary->{ZERO_INT_LIBS}, 'external', 'library', 'libraries'), 'libs', '&filt_show=unneeded' )
			.' <b>without using interfaces</b> in '.(($summary->{ZERO_INT_LIBS} > 1) ? 'those libraries' : 'this library').'.'
			.' The Application would be more portable if it is fixed to not use unneeded libraries.';
		$res .= '</p>';
	}
	
	# Unapproved libraries
	if ( $summary->{UNAPPROVED_LIBS} ) {
		$res .= '<p>';
		$res .= $warn_image; # Yellow image
		$res .= 'The Application uses '
			.refer( _pl($summary->{UNAPPROVED_LIBS}, 'external', 'library', 'libraries'), 'libs', '&filt_show=uncommon' )
			.' that '.($summary->{UNAPPROVED_LIBS}==1?'is':'are').' <b>not common</b> in distributions.'
			.' Please consider providing such libraries as a part of the Application package.';
		$res .= '</p>';
	}

	if ( $summary->{NO_SHT_HASH} ) {
		$res .= '<p>';
		$res .= $warn_image; # Yellow image
		if ( $summary->{NO_SHT_HASH} == 1 ) { # only one
			$res .= 'There is '.refer( '<b>1</b> application component', 'com', '' ).' which has <b>no `.hash\' section</b>. This component would fail to run on older distributions (e.g., RHEL4).';
		} else { # plural
			$res .= 'There are '.refer( '<b>'.$summary->{NO_SHT_HASH}.'</b> application components', 'com', '' )
				.' that have <b>no \'.hash\' section</b>. These components will fail to run on older distributions (e.g., RHEL4).';
		}
		$res .= '</p>';
	}


	$res .= "$nl</div>$nl<br />$nl";
	$res .= <<DATA;

<div id="tooltip" style="display: none; background-color: #ffffd0; border: solid 1px black; position: absolute; z-index: 1000;">
<table cellpadding="0" cellspacing="0" border="0" style="width: 17em;">
  <tr style="background-color: #d0d0d0; color: black;">
    <td align="right" style="border-bottom: solid 1px black;"><img src="images/environment-spacer.gif" width="7" height="1" alt="" /></td>
    <td align="center" width="100%" style="border-bottom: solid 1px black;"><b>Information</b></td>
    <td align="right" style="border-bottom: solid 1px black;"><img src="images/close.png" onclick="javascript:initDialog(null);" width="20" height="20" title="Close" alt="X" style="cursor: pointer;" /></td>
  </tr>
  <tr>
    <td colspan="3" style="padding: 5px; white-space: nowrap;" id="tooltip_text"></td>
  </tr>
</table>
</div>

<table cellpadding="0" cellspacing="0" border="0" style="display: none; position: absolute; z-index: 999;" id="shadow">
  <tr>
    <td width="14" height="14" style="background: url('images/shadow_tl.png') no-repeat;"></td>
    <td height="14" style="background: url('images/shadow_tc.png') repeat-x;"></td>
    <td width="14" height="14" style="background: url('images/shadow_tr.png') no-repeat;"></td>
  </tr>
  <tr>
    <td width="14" style="background: url('images/shadow_ml.png') repeat-y;"></td>
    <td style="background: url('images/shadow_mc.png') repeat;" id="shadow_center"></td>
    <td width="14" style="background: url('images/shadow_mr.png') repeat-y;"></td>
  </tr>
  <tr>
    <td width="14" height="14" style="background: url('images/shadow_bl.png') no-repeat;"></td>
    <td height="14" style="background: url('images/shadow_bc.png') repeat-x;"></td>
    <td width="14" height="14" style="background: url('images/shadow_br.png') no-repeat;"></td>
  </tr>
</table>

<script type="text/javascript" language="javascript">
// <![CDATA[

preload_images(new Array('status_collapse.png', 'close.png', 'shadow_bc.png', 'shadow_bl.png', 'shadow_br.png', 'shadow_mc.png', 'shadow_ml.png', 'shadow_mr.png', 'shadow_tc.png', 'shadow_tl.png', 'shadow_tr.png'));

var just_shown_dialog = false;          // Indicates that the dialog was just shown (to prevent its hiding from onBodyClick)
var dialog = null;                      // The dialog itself
var dialog_caller = null;               // Element which caused the dialog to appear
var dialog_caller_class = null;         // Normal class name of the dialog caller (to be restored when the dialog is closed)
var dialog_shadow = document.getElementById('shadow');
var dialog_shadow_center = document.getElementById('shadow_center');

// Hides the previous dialog (if it was shown) and prepare data for
// displaying the new one.
// CALLER - the HTML element which caused the dialog to appear and (possibly)
// will change its appearance because of that.
function initDialog(caller) {
	if (dialog_caller && dialog_caller_class)
		dialog_caller.className = dialog_caller_class;
	dialog_shadow.style.display = 'none';
	if (dialog)
		dialog.style.display = 'none';
	if (!caller || (caller == dialog_caller)) {
		dialog_caller = null;
		dialog_caller_class = null;
		return false;
	}
	else {
		dialog_caller = caller;
		dialog_caller_class = caller.className;
		return true;
	}
}

// Shows the dialog at the specified position and adjusts the shadow below it.
function showDialog(show_dlg, pos) {
	dialog = show_dlg;
	dialog.style.left = pos.x + 'px';
	dialog.style.top = pos.y + 'px';
	dialog.style.display = '';
	just_shown_dialog = true;
	dialog_shadow_center.width = dialog.offsetWidth - 15;
	dialog_shadow_center.height = dialog.offsetHeight - 15;
	dialog_shadow.style.left = pos.x + 'px';
	dialog_shadow.style.top = pos.y + 'px';
	dialog_shadow.style.display = '';
}

function onBodyClick() {
	if (!just_shown_dialog)
		initDialog(null);
	just_shown_dialog = false;
}

// Tooltip processing functions
var tooltip_texts = new Array();
tooltip_texts['libs_ints_common'] = 'Common libraries are manually selected by the Linux Foundation<br />to represent .so names of those libraries that may be externally required<br />by applications and provided by distributions.<br />Please write to <a href="mailto:lsb-appcheck-support\@linuxfoundation.org">lsb-appcheck-support\@linuxfoundation.org</a> to suggest<br />adding more libraries to the "common libraries" list.';
tooltip_texts['libs_ints_unknown'] = 'Unknown libraries are those not included in the "approved" list<br />of common libraries. This means such libraries are not supposed<br />to be externally required by applications and correspondingly provided<br />by distributions. Sometimes new versions of common libraries<br />fall into the "unknown" category as well. It is recommended to include<br />such libraries as a part of your application (either by statically linking<br />or by supplying the .so in your package).';
tooltip_texts['ints_linkage_failed'] = 'The analyzer has failed to determine which library (among<br />those required by this application component) provides this<br />interface. This may happen if the corresponding library<br />wasn\\\'t found in the database or in the distribution on this<br />machine, or if the list of libraries needed by this application<br />component is incomplete. If this binary is executable, then<br />most likely it will fail if launched on this machine.';
tooltip_texts['com_common_lib'] = 'This library is provided by some distributions.<br />Depending on which distributions provide this library, you can<br />consider excluding this library from the application package<br />(making it an external dependency), thus reducing the<br />package size.';
tooltip_texts['com_no_hash_section'] = 'The ELF binaries have a hash section for runtime symbol<br/>lookup to be used by the dynamic linker. This is the \\'.hash\\'<br/>section in the old \\'sysv\\' format and the \\'.gnu.hash\\' section<br/>in the new \\'gnu\\' format. The latter format is up to 50\%<br/>more effective. In newer distributions, the dynamic linker<br/>can handle both types of hash formats, while older<br/>distributions can only work with the older \\'sysv\\' format<br/>and therefore require a \\'.hash\\' ELF section.<br/>Which section is present in the binary is determined by<br/>the linker switch \\'--hash-style\\' at compile time.';

function show_hide_tooltip(elem_parent, id, event) {
	if (!initDialog(elem_parent))
		return;
	var tt_window = document.getElementById('tooltip');
	if (tt_window) {
		var tt_text = document.getElementById('tooltip_text');
		if (tt_text) {
			if (tooltip_texts[id])
				tt_text.innerHTML = tooltip_texts[id];
			else
				tt_text.innerHTML = '<font color="red"><b>Sorry, no information for this item was found!</b></font><br />Please write to <a href="mailto:lsb-appcheck-support\@linuxfoundation.org">lsb-appcheck-support\@linuxfoundation.org</a>.';
			var scrollTop = document.body.scrollTop;
			if (scrollTop == 0)
			{
			    if (window.pageYOffset)
			        scrollTop = window.pageYOffset;
			    else
			        scrollTop = (document.body.parentElement) ? document.body.parentElement.scrollTop : 0;
			}
			var scrollLeft = document.body.scrollLeft;
			if (scrollLeft == 0)
			{
			    if (window.pageXOffset)
			        scrollLeft = window.pageXOffset;
			    else
			        scrollLeft = (document.body.parentElement) ? document.body.parentElement.scrollLeft : 0;
			}
			var div_top = event.clientY + scrollTop + 10;
			var div_left = event.clientX + scrollLeft + 10;
			tt_window.style.display = '';
			var div_width = (tt_window.style.pixelWidth || tt_window.offsetWidth);
			var wnd_width = (document.body.clientWidth || window.innerWidth);
			if (div_left + div_width > wnd_width + scrollLeft)
				div_left = wnd_width + scrollLeft - div_width;
			if (div_left < scrollLeft)
				div_left = scrollLeft;
			showDialog(tt_window, {x: div_left, y: div_top});
		}
	}
}

// ]]>
</script>
DATA

	return $res;
}

sub ints_data {
	my ($filt_distr, $filt_com, $filt_lib) = @_;
	
	my $ints = {};
	# $ints->{$Iname}{$ver} = {
	#   VER_FOUND => $ver_found,
	#   UNAPPROVED,
	#   DISTRS => {$Did => 1},
	#   LIBS => {$lib_soname} => {
	#      COMPONENTS => {$com => 1},
	#      LSB_STATUS, MORE_INFO
	#   }
	# };
	
	my $nonfilt = { INTS => {}, LIBS => {} }; # Non-filtered information
	
	foreach my $com_obj ( values %$Analysis::app_components ) {
		# Non-filtered information
		foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
			$nonfilt->{INTS}{$fullname} = 1;
		}
		
		foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
			my ($Iname, $ver) = split /\@/, $fullname;
			$ver = "" if !defined $ver;
			
			my $int = $ints->{$Iname}{$ver} ||= { DISTRS=>{}, LIBS=>{} };
			
			foreach my $Did ( keys %$Analysis::all_distrs ) {
				my $tmp;
				if ( $tmp = $com_obj->{$Did}{FOUND_INTS}{$fullname} ) {
					my ($soname, $ver_found) = split /\|/, $tmp;
					
					my $lib_obj = $com_obj->{$Did}{LIBS}{$soname};
					if ( $lib_obj && $lib_obj->{VIA} ) { $soname = $lib_obj->{VIA}; }
					
					# Non-filtered information
					$nonfilt->{LIBS}{$soname} = 1;
					
					if ( $filt_lib && $soname ne $filt_lib ) { next; } # Library filter
					if ( $filt_distr && !$filt_distr->{$Did} ) { next; } # Distribution filter
					if ( $filt_com && $com_obj->{NAME} ne $filt_com ) { next; } # Component filter
					
					$int->{VER_FOUND} = $ver_found if $ver_found; # TODO: different VER_FOUND's (on different distros)
					$int->{LIBS}{$soname}{COMPONENTS}{$com_obj->{NAME}} = 1;
					unless ( defined $int->{DISTRS}{$Did} ) { # do not set 1 if there is already 0 or 1.
						$int->{DISTRS}{$Did} = 1;
					}
				}
				elsif ( $tmp = $com_obj->{$Did}{MISSING_INTS}{$fullname} ) {
					my $soname = $tmp;
					
					# Non-filtered information
					$nonfilt->{LIBS}{$soname} = 1;
					
					if ( $filt_lib && $soname ne $filt_lib ) { next; } # Library filter
					if ( $filt_distr && !$filt_distr->{$Did} ) { next; } # Distribution filter
					if ( $filt_com && $com_obj->{NAME} ne $filt_com ) { next; } # Component filter
					
					$int->{LIBS}{$soname}{COMPONENTS}{$com_obj->{NAME}} = 1;
					$int->{DISTRS}{$Did} = 0;
				}
				else { # Non-linked for this distro
					my $soname = ""; # library for this interface wasn't found.
					
					# Non-filtered information
					$nonfilt->{LIBS}{"~"} = 1; # symbol '~' goes after a-z in alphabetical order
					
					if ( $filt_lib && $filt_lib ne 'nonlinked' ) { next; } # Library filter
					if ( $filt_distr && !$filt_distr->{$Did} ) { next; } # Distribution filter
					if ( $filt_com && $com_obj->{NAME} ne $filt_com ) { next; } # Component filter
					
					$int->{LIBS}{$soname}{COMPONENTS}{$com_obj->{NAME}} = 1;
					$int->{DISTRS}{$Did} = 0;
				}
			}
		}
	}
	
	foreach my $Iname ( keys %$ints ) {
	foreach my $ver ( keys %{$ints->{$Iname}} ) {
		my $int = $ints->{$Iname}{$ver};
		foreach my $soname ( keys %{$ints->{$Iname}{$ver}{LIBS}} ) { # Look for unapproved libs
			if ( $soname && !$Analysis::approved_libs->{$soname} ) {
				$int->{UNAPPROVED} = 1;
				last;
			}
			if ( !$soname && $Analysis::unknown_ints->{$Iname} ) {
				$int->{UNAPPROVED} = 1;
			}
		}
		
		# Delete distributions where the interface is missing
		foreach my $Did ( grep {!$int->{DISTRS}{$_}} keys %{$int->{DISTRS}} ) {
			delete $int->{DISTRS}{$Did};
		}
	}}
	
	if ( wantarray ) {
		return ($ints, $nonfilt);
	}
	return $ints;
}

sub ints_lsb_data {
	my ($ints, $selected_lsb_ver) = @_;
	
	foreach my $Iname ( keys %$ints ) {
	foreach my $ver ( keys %{$ints->{$Iname}} ) {
	foreach my $soname ( keys %{$ints->{$Iname}{$ver}{LIBS}} ) {
		my $int_obj = $ints->{$Iname}{$ver}{LIBS}{$soname};
		
		# LSB Status
		$int_obj->{LSB_STATUS} = Analysis::int_lsb_status($Iname, $ver, $soname, $selected_lsb_ver);
		
		if ( $Analysis::int_attrs->{$Iname}
			&& $Analysis::int_attrs->{$Iname}{Analysis::soname_to_libname($soname)} ) 
		{
			# MORE_INFO
			$int_obj->{MORE_INFO} = 'MORE_INFO';
		}
	}}}
}

# Library compatibility table
sub I_tab {
	my ($selected_lsb_ver, $distr_sel, $filt_com, $filt_lib, $filt_show) = @_;

	my $res = "";
	my $nl = "\n";

	my $filtered = $filt_com || $filt_lib || $filt_show || scalar grep {!$_} values %$distr_sel;
	
	my ($ints, $nonfilt_info) = ints_data( $distr_sel, $filt_com, $filt_lib );
	ints_lsb_data($ints, $selected_lsb_ver);
	
	# Check if the table is empty
	if ( !$filtered && !%$Analysis::app_components ) {
		$res .= no_binary_div();
		return $res;
	}
	if ( !$filtered && !%$ints ) {
		$res .= no_binary_div("The application requires no external interfaces.");
		return $res;
	}

	$res .= <<HEREDOC;
The table below lists all external interfaces required by your Application (based on ELF symbols). LSB and Distribution presence statuses are provided for each interface.<br /><br />
HEREDOC

	my $link = href( tab_link('ints') );
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id )
{
	var v = document.getElementById( id );
	if ( v ) {
		v.style.display = (v.style.display == 'none' ? '' : 'none');
	}
	return false;
}

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes'
		);
	w.moveTo(100,0);
	w.focus();

	return false;
}
// ]]>
</script>
HEREDOC
	
	my ($table, $row_n) = I_table($ints, $selected_lsb_ver, $distr_sel, $filt_com, $filt_lib, $filt_show);
	
	# Filters
	my $filt = "";
	
	$filt .= distr_selector( $distr_sel ); # Distribution selector
	
	$filt .= '<form name="filter" method="get" action="">';
	$filt .= '<table><tr><td>';
	
	$filt .= '<table cellspacing="0" class="bartable"><tr>';
	
	# "Component" filter
	$filt .= '<td style="border-right:0"><b>Component:</b></td>'
		.'<td>'
		.'<select name="filt_com" id="filt_com" onchange="onFilter()" style="width: 20em;">'.$nl
		.'<option value="">All</option>'.$nl;
	foreach my $com ( sort keys %$Analysis::app_components ) {
		my ($disp_name, $disp_path) = parse_app_component($com);
		my $selected = "";
		if ( $filt_com && $com eq $filt_com ) {
			$selected = ' selected="selected"';
		}
		$filt .= '<option value="'.html($com).'"'
					.$selected.'>'.html($disp_name).'</option>'.$nl;
	}
	$filt .= '</select>'.$nl
		.'</td>'; # </filt_com>

	# spacer
	$filt .= '<td style="padding:0.2em;border:0;background:white"></td>';

	# "Library" filter
	$filt .= '<td style="border-right:0"><b>Library:</b></td>'
		.'<td>'
		.'<select name="filt_lib" id="filt_lib" onchange="onFilter();" style="width: 12.5em;">'.$nl
		.'<option value="">All</option>'.$nl;
	foreach my $soname ( sort {lc $a cmp lc $b} keys %{$nonfilt_info->{LIBS}} ) {
		my $val = $soname;
		if ( $val eq "~" ) { $val = "nonlinked"; }
		my $display = $val;
		if ( $display eq "nonlinked" ) { $display = "Non-linked"; }
		my $selected = "";
		if ( $filt_lib && $val eq $filt_lib ) {
			$selected = ' selected="selected"';
		}
		$filt .= '<option value="'.html($val).'"'
					.$selected.'>'.html($display).'</option>'.$nl;
	}
	$filt .= '</select>'.$nl
		.'</td>'; # </filt_lib>
	
	# spacer
	$filt .= '<td style="padding:0.2em;border:0;background:white"></td>';
	
	# "Show" filter
	$filt .= '<td style="border-right:0"><b>Show:</b></td>'
		.'<td>'
		.'<select name="filt_show" id="filt_show" onchange="onFilter()" style="width: 12.5em;">'.$nl;
	my @filt_show_values = ( ""=>"All", "problems"=>"Compatibility problems",
			"non-lsb"=>"Non-LSB", "deprecated"=>"Deprecated" );
	while ( @filt_show_values ) {
		my $key = shift @filt_show_values;
		my $val = shift @filt_show_values;
		my $selected = "";
		if ( $filt_show && $key eq $filt_show ) {
			$selected = ' selected="selected"'; 
		}
		$filt .= '<option value="'.$key.'"'
					.$selected.'>'.$val.'</option>'.$nl;
	}
	$filt .= '</select>'.$nl
		.'</td>'; # </filt_show>
		
	$filt .= '</tr></table>';
	$filt .= '</td>';

	# Filter status
	my $all_n = scalar keys %{$nonfilt_info->{INTS}};
	$filt .= '<td style="padding-left:20pt">';
	$filt .= '<span id="lines_count">';
	if ( $filtered ) {
		$filt .= '<span style="color:red"><b>'.$row_n.'</b></span> of ';
	}
	$filt .= '</span>';
	$filt .= $all_n.' interface'.($all_n!=1?'s':'').' shown.';
	
	$filt .= '</td></tr></table>'; # </filters>
	$filt .= '</form>';
	
	## Ajax stuff
	$res .= show_error_dlg(''); # placeholder for an error box
	
	$res .= ajax_wait(); # ajax hourglass
	
	my $ajax_link = href('details='.$_GET{'details'});
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[

timeout_show_progress = 0;

var cur_ifc = '$filt_com';
var cur_ifl = '$filt_lib';
var cur_ifs = '$filt_show';

function onFilter() {
	var filt_com = document.getElementById('filt_com');
	var filt_lib = document.getElementById('filt_lib');
	var filt_show = document.getElementById('filt_show');
	
	if ( filt_com ) {
		cur_ifc = filt_com.value;
		setCookie('ifc', cur_ifc, null);
	}
	if ( filt_lib ) {
		cur_ifl = filt_lib.value;
		setCookie('ifl', cur_ifl, null);
	}
	if ( filt_show ) {
		cur_ifs = filt_show.value;
		setCookie('ifs', cur_ifs, null);
	}
	
	if (!ajax_req_busy) {
		ajax_call_get('action=results_table&table=ints&${ajax_link}');
	}
	
	return false;
}

function ajaxProcessResult(responseXML) {
	
	var i_table_html;
	if (responseXML.getElementsByTagName('i_table').length > 0) {
		i_table_html = responseXML.getElementsByTagName('i_table')[0].childNodes[0].nodeValue;
	}
	
	var i_table_wrap = document.getElementById('i_table_wrap');
	
	if ( i_table_html && i_table_wrap ) {
		i_table_wrap.innerHTML = i_table_html;
	} else {
		return -1;
	}
	
	var lines_count = document.getElementById('lines_count');
	if ( lines_count ) {
		var row_n;
		if (responseXML.getElementsByTagName('row_n').length > 0) {
			row_n = responseXML.getElementsByTagName('row_n')[0].childNodes[0].nodeValue;
			lines_count.innerHTML = '<span style="color:red"><b>'+row_n+'</b></span> of&nbsp;';
		} else {
			lines_count.innerHTML = '';
		}
	}
}

// Cookies on focus:
if (typeof(document.addEventListener) != 'undefined') {
 document.addEventListener('focus', doFocus, false); 
} else if (typeof(window.attachEvent) != 'undefined') {
 window.attachEvent('onfocus', doFocus);
}

function doFocus(event) {
	setCookie('ds', cur_ds, null);
	setCookie('ifc', cur_ifc, null);
	setCookie('ifl', cur_ifl, null);
	setCookie('ifs', cur_ifs, null);
}

// ]]>
</script>
HEREDOC
	
	## Put the filters
	$res .= $filt;
	
	## Put the table
	$res .= '<div id="i_table_wrap">'.$nl;
	$res .= $table;
	$res .= '</div>'.$nl;
	##
	
	return $res;
}

sub I_table {
	my ($ints, $selected_lsb_ver, $distr_sel, $filt_com, $filt_lib, $filt_show) = @_;
	
	my $filt_distr = scalar grep {!$_} values %$distr_sel;
	my $filtered = $filt_com || $filt_lib || $filt_show || $filt_distr;
	
	my $table = "";
	my $nl = "\n";
	
	# Table Header
	my $table_open = '<table id="i_table" class="wikitable" width="100%">';
	$table .= $table_open.$nl;
	$table .= '<tr>'
		.'<th>Interface</th>'
		.'<th>Version</th>'
		.'<th>Dependent<br/>Application&nbsp;Components</th>'
		.'<th>Library</th>'
		.'<th>Presence in<br />'.($filt_distr ? 'Selected ' : '').'Distributions</th>'
		.'<th id="lsb_col_header" style="white-space:nowrap">In&nbsp;LSB&nbsp;'.$selected_lsb_ver.'?</th>'
		.'<th>More Info</th>'
		.'</tr>';

	my $row_n = 0;
	my $aux_id = 0;
	my $interface_n = 0;
	
	my %name_sort = ();
	foreach my $Iname ( keys %$ints ) {
		$name_sort{$Iname} = lc( $Analysis::unmangled->{$Iname} || $Iname );
	}
	
	# Table Body
	my $prev = "";
	foreach my $Iname ( sort {$name_sort{$a} cmp $name_sort{$b}} keys %$ints ) {
	foreach my $ver ( sort keys %{$ints->{$Iname}} ) {
		my $int = $ints->{$Iname}{$ver};
		
		my $rowspan = 0;
		my @col_LSB = ();
		my @col_COM = ();
		my @col_LIB = ();
		my @col_INF = ();
		foreach my $soname ( sort keys %{$ints->{$Iname}{$ver}{LIBS}} ) {
			my $int_obj = $ints->{$Iname}{$ver}{LIBS}{$soname};
			
			## LSB
			my $col_lsb = "";

			my $lsb_status = $int_obj->{LSB_STATUS}{STATUS};
			my $lsb_status_ver = $int_obj->{LSB_STATUS}{STATUS_VER};
			
			my $color = undef;
			if ( $lsb_status eq "NOT" || $lsb_status eq "WITHDRAWN" ) {
				my $txt = 'No';
				my $color = 'red';
				if ( $int_obj->{LSB_STATUS}{REASON} ) {
					my $reason = $int_obj->{LSB_STATUS}{REASON};
					if ( $reason eq 'WRONG_VER' ) {
						$txt = 'Wrong&nbsp;ver';
						$color = '#A04020';
					}
					elsif ( $reason eq 'WRONG_LIB' ) {
						$txt = 'Wrong&nbsp;lib';
						$color = '#A04020';
					}
				}
				$col_lsb .= '<span style="color:'.$color.'">'.$txt.'</span>';
			}
			elsif ( $lsb_status eq "CANDIDATE" ) {
				$col_lsb .= '<span style="color:#907000">Candidate for '.$lsb_status_ver.'</span>';
			}
			elsif ( $lsb_status eq "FUTURE" ) {
				$col_lsb .= '<span style="color:#907000">In LSB '.$lsb_status_ver.'</span>';
			}
			elsif ( $lsb_status eq "OPTIONAL" ) {
				if ( $filt_show && $filt_show eq 'non-lsb' ) { next; }
				$col_lsb .= '<span style="color:blue">Optional</span>';
			}
			elsif ( $lsb_status eq "IN" ) {
				if ( $filt_show && $filt_show eq 'non-lsb' ) { next; }
				$col_lsb .= '<span style="color:green">Yes</span>';
			}
			elsif ( $lsb_status eq "DEPRECATED" ) {
				if ( $filt_show && $filt_show eq 'non-lsb' ) { next; }
				$col_lsb .= '<span style="color:blue">Deprecated</span>';
			}
			if ( $lsb_status ne "DEPRECATED" ) {
				if ( $filt_show && $filt_show eq 'deprecated' ) { next; }
			}
			$col_LSB[$rowspan] = $col_lsb;
			
			## App modules expandable list
			my $app_components = "";
			my $app_comp_n = scalar keys %{$int_obj->{COMPONENTS}};
			foreach my $com ( sort keys %{$int_obj->{COMPONENTS}} ) {
				my ($disp_name, $disp_path) = parse_app_component($com);
				
				$app_components .= '<div class="indentwrap" title="'.html($disp_path).'">'.html($disp_name).'</div>';
			}
			my $div_id = 'cl'.($aux_id++);
			my $col_components = "";
			$col_components .= '<a href="" onclick="return toggle(\''.$div_id.'\')">'
						.$app_comp_n.' component'.( $app_comp_n > 1 ? 's' : '' )
						.' (&hellip;)'
					.'</a>';
			$col_components .= '<div id="'.$div_id.'" style="display:none">'.$nl;
			$col_components .= $app_components;
			$col_components .= '</div>';
			
			$col_COM[$rowspan] = $col_components;

			## Library
			my $col_library = "";
			if ( $soname eq "" ) {
				$col_library = '<span style="color:red">Non-linked</span>';
				# Tooltip:
				$col_library .= '&nbsp;<a href="tests_help.pl#ints_linkage_failed" onclick="javascript:show_hide_tooltip(this, \'ints_linkage_failed\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
			}
			else {
				my $soname_link = href( 'http://www.linuxfoundation.org/navigator/browse/rawlib.php'
						.'?cmd=list-byrunname&RLrunname='.$soname );
				if ( !$Analysis::approved_libs->{$soname} ) {
					$col_library = '<span style="color:#907000">'.html($soname).'</span>';
					# Tooltip:
					$col_library .= '&nbsp;<a href="tests_help.pl#libs_ints_unknown" onclick="javascript:show_hide_tooltip(this, \'libs_ints_unknown\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
				} else {
					#$col_library = '<a href="'.html($soname_link).'" target="_blank"'
							#.($color ? ' style="color:'.$color.'"' : '').'>'.html($soname).'</a>';
					$col_library = html($soname);
				}
			}
			$col_LIB[$rowspan] = $col_library;
			
			## More Info
			my $col_moreinfo = "";
			
			my $info_link = href( tab_link('ints').'&int='.$Iname.' '.$ver.' '.$soname );
			my $info_onclick = 'onclick="return popup_win(\''.html($info_link).'\')"';
			my $txt = 'Details&hellip;';
			if ( $int_obj->{MORE_INFO} && $int_obj->{MORE_INFO} eq 'MORE_INFO' ) {
				$txt = '<b>More&nbsp;Info&hellip;</b>';
			}
			$col_moreinfo .= '<a href="'.html($info_link).'" '.$info_onclick.'>'.$txt.'</a>';
			
			$col_INF[$rowspan] = $col_moreinfo;
			
			$rowspan++;
		}
		
		next if $rowspan == 0; # Filtered out
		
		## Presence in distributions
		my $total = scalar grep {$_} values %$distr_sel;
		my $present = 0;
		if ( $int->{DISTRS} ) {
			$present = scalar keys %{$int->{DISTRS}};
		}

		my $bgcolor;
		my $col_distrs = "";
		my $pr_link = href( tab_link('ints').'&list='.$Iname.' '.$ver );
		my $pr_onclick = 'onclick="return popup_win(\''.html($pr_link).'\')"';
		if ( $present == $total ) {
			if ( $filt_show && $filt_show eq 'problems' ) { next; }
			$bgcolor = "#CCFFCC"; # green
			$col_distrs .= $present." of ".$total;
			$col_distrs .= ' (<a href="'.html($pr_link).'" '.$pr_onclick.'>list&hellip;</a>)';
		} else {
			if ( $present > 0 ) {
				$bgcolor = "#FEFECC"; # yellow
				$col_distrs .= $present." of ".$total;
				if ( $int->{UNAPPROVED} ) {
					$col_distrs .= ' known';
				}
				$col_distrs .= ' (<a href="'.html($pr_link).'" '.$pr_onclick.'>list&hellip;</a>)';
			} else {
				if ( $int->{UNAPPROVED} ) {
					$bgcolor = "#FEECCC"; # dark yellow
					$col_distrs .= 'Unknown';
				} else {
					$bgcolor = "#FFCCCC"; # red
					$col_distrs .= "Missing";
				}
			}
		}
		
		# wrap unmangled name (at _) if it is too long
		my $tmp = "";
		my $unmangled = $Analysis::unmangled->{$Iname} || $Iname;
		my $displayname = $unmangled;
		$displayname =~ s/\(/ \(/g;
		$displayname =~ s/\)/\) /g;
		while ( $displayname =~ s/^(\S{0,40}?\s+|\S{30}[^_]{0,20}|\S{1,40}$)// ) {
			$tmp .= html($1);
			$tmp .= "<br/>" if $displayname && $1 !~ /\s$/;
		}
		$displayname = $tmp;
		
		my $iname_link =  href("http://www.linuxfoundation.org/navigator/browse/int_single.php"
			."?cmd=list-by-name&Iname=".$Iname);
		
		if ( keys %{$ints->{$Iname}{$ver}{LIBS}} == 1 ) {
			my ($soname) = keys %{$ints->{$Iname}{$ver}{LIBS}};
			$iname_link .= '&Ilibrary='.Analysis::soname_to_libname($soname);
		}
		
		## Interface name
		my $col_iname = '<a href="'.html($iname_link).'" target="_blank"'
				.'>'.$displayname.'</a>'; # TODO: css class to make it look like foreign link
		
		## Version
		my $col_ver = $ver;
		if ( $int->{VER_FOUND} ) {
			$col_ver .= ' ('.$int->{VER_FOUND}.')';
		}
		
		for ( my $n = 0; $n < $rowspan; $n++ ) {
			$table .= '<tr>'.$nl;
			
			# Interface name column
			if ( $n == 0 ) {
				$table .= '<td class="indentwrap"'.($rowspan>1 ? ' rowspan="'.$rowspan.'"' : '' ).'>'.$nl;
				$table .= $col_iname;
				$table .= '</td>'.$nl;
			}
			
			# Version column
			if ( $n == 0 ) {
				$table .= '<td '.($rowspan>1 ? ' rowspan="'.$rowspan.'"' : '' ).'>';
				$table .= $col_ver;
				$table .= '</td>'.$nl;
			}
			
			$table .= '<td align="left">';
			$table .= $col_COM[$n];
			$table .= '</td>'.$nl;
			
			$table .= '<td style="white-space:nowrap">';
			$table .= $col_LIB[$n];
			$table .= '</td>'.$nl;
			
			if ( $n == 0 ) {
				$table .= '<td '.($rowspan>1 ? ' rowspan="'.$rowspan.'"' : '' )
					.' align="right" style="background:'.$bgcolor.';white-space:nowrap">';
				$table .= $col_distrs;
				$table .= '</td>'.$nl;
			}
			
			$table .= '<td align="center">';
			$table .= $col_LSB[$n];
			$table .= '</td>'.$nl;
			
			$table .= '<td>';
			$table .= $col_INF[$n];
			$table .= '</td>'.$nl;
			
			$table .= '</tr>'.$nl;
			$row_n++;
		}
		$interface_n++;
	}}
	
	# Table Footer
	$table .= '</table>'.$nl;
	
	if ( $filtered && $row_n == 0 ) {
		$table .= '<div style="padding:10pt;text-align:center;font-size:large;color:#009">'
					.'<p>No interfaces match the filter.</p>'
				.'</div>';
	}
	
	return ($table, $interface_n);
}

sub int_details {
	my ($Iname, $ver, $soname, $selected_lsb_ver) = @_;
	$ver ||= ""; $soname ||= "";
	
	my $nl = "\n";
	my $res = "";
	
	my $ints = ints_data();
	ints_lsb_data($ints, $selected_lsb_ver);
	
	my $int_obj = $ints->{$Iname}{$ver}{LIBS}{$soname}
		or return "Error: There is no such interface: '$Iname' '$ver' '$soname'";
	
	my $unmangled = $Analysis::unmangled->{$Iname} || $Iname;
	
	# Reference to the LSB Navigator:
	my $navi_link = href("http://www.linuxfoundation.org/navigator/browse/int_single.php"
			."?cmd=list-by-name&Iname=".$Iname);
	if ( $soname ) {
		my $Lname = $Analysis::lsb_libs->{$soname}{LSBNAME};
		if ( !$Lname ) {
			$Lname = Analysis::soname_to_libname($soname);
		}
		if ( $Lname eq 'libstdc++' ) { $Lname = 'libstdcxx'; }
		$navi_link .= href('&Ilibrary='.$Lname);
	}

	# Title
	$res .= '<h1>Interface: <a href="'.html($navi_link).'" target="_blank">'.html($unmangled).'</a>'
			.' ('.($ver ? html($ver).', ' : '')
				.($soname ? '<a href="'.html(href(tab_link('libs').'&lib='.$soname)).'">'.html($soname).'</a>'
				          : '<span style="color:red">non-linked</span>')
			.')'.'</h1>';
	if ( $Analysis::unmangled->{$Iname} ) {
		$res .= '<p><b>C++ mangled name:</b> '.html($Iname).'</i></p>';
	}
	$res .= '<hr />'.$nl;
	
	# Int KB info still not implemented
	
	# Check IntAttr
	my $attr = $Analysis::int_attrs->{$Iname}{Analysis::soname_to_libname($soname)};
	if ( $attr ) {
		foreach my $type ( sort keys %$attr ) {
			my $attr_html = "";
			my $bgcolor = '#fec';
			if ( $type eq "Recommendation" ) {
				$attr_html .= '<h3>LSB Recommendation</h3>';
				$attr_html .= $attr->{$type};
			}
			elsif ( $type eq "Rejection" ) {
				$attr_html .= '<h3>LSB Rejection</h3>';
				$attr_html .= $attr->{$type};
			}
			else { # Custom
				$attr_html .= $attr->{$type};
			}
			$attr_html =~ s/^\s+//; # remove unnecessary spaces
			$attr_html =~ s/\s+$//; #
			$attr_html .= "." if $attr_html =~ /[^\s\.]\s*$/;
			
			$res .= '<div class="box" style="background:'.$bgcolor.'">'.$attr_html.'</div>';
		}

		$res .= '<hr />'.$nl;
	}

	# LSB Status
	$res .= '<h2>LSB Status:</h2>'.$nl;

	my $lsb_status_obj = $int_obj->{LSB_STATUS};
	my ($lsb_status, $lsb_status_ver) = @$lsb_status_obj{qw/STATUS STATUS_VER/};
	
	if ( $lsb_status eq "NOT" ) {
		if ( $lsb_status_obj->{REASON} && $lsb_status_obj->{REASON} eq 'WRONG_LIB' ) {
			$res .= '<div style="color:red">This interface is included in LSB, but assigned to another library. Click the interface name above for details.</div>'.$nl;
			# TODO: ^ what library
		}
		elsif ( $lsb_status_obj->{REASON} && $lsb_status_obj->{REASON} eq 'WRONG_VER' ) {
			$res .= '<div style="color:red">Another version of the interface is included in LSB.  Click the interface name above for details.</div>'.$nl;
			# TODO: ^ what version
		}
		else {
			$res .= '<div style="color:red">Never been in LSB.</div>'.$nl;
		}
	}
	elsif ( $lsb_status eq "FUTURE" ) {
		$res .= '<div style="color:red">Not in LSB '.$selected_lsb_ver.'. (Appeared in LSB '.$lsb_status_ver.'.)</div>'.$nl;
	}
	elsif ( $lsb_status eq "CANDIDATE" ) {
		$res .= '<div style="color:red">Not in LSB. Candidate for LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "OPTIONAL" ) {
		$res .= '<div style="color:blue">Optional since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "IN" ) {
		$res .= '<div style="color:green">Mandatory since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "DEPRECATED" ) {
		$res .= '<div style="color:blue">Deprecated since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	elsif ( $lsb_status eq "WITHDRAWN" ) {
		$res .= '<div style="color:red">Withdrawn since LSB '.$lsb_status_ver.'.</div>'.$nl;
	}
	
	#if ( $soname ) {
		#my $libname = $Analysis::lsb_libs->{$soname}{LSBNAME};
		#if ( $libname ) {
			#my $link = href('http://www.linuxfoundation.org/navigator/browse/lib_single.php'
					#.'?cmd=list-by-name&Lname='.$libname);
			#$res .= '<p>This interface is assigned to '
				#.'<a href="'.html($link).'" target="_blank"><b>'.html($libname).'</b></a>'
				#.' library.</p>';
		#}
	#}
	
	$res .= '<hr />'.$nl;
	
	$res .= '<h2>This interface is used by following aplication components:</h2>';
	
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id, obj )
{
	var v = document.getElementById( id );
	if ( v ) {
		if ( v.style.display == 'none' ) {
			v.style.display = '';
			if ( obj ) obj.innerHTML = '&ndash;';
		} else {
			v.style.display = 'none';
			if ( obj ) obj.innerHTML = '+';
		}
	}
	return false;
}
// ]]>
</script>
HEREDOC

	local *distr_list = sub {
		my ($color, $com, %distrs) = @_;

		my $res = "";

		my $all_distrs = $Analysis::all_distrs;
		my @distrs = sort {lc $all_distrs->{$a}{FULLNAME} cmp lc $all_distrs->{$b}{FULLNAME}} keys %distrs;
		$res .= '<ul>'.$nl;
		for my $Did ( @distrs ) {
			my $link = href(tab_link('com').'&distr='.$Did.'&com='.$com);
			
			$res .= '<li>';
			$res .= '<a style="color:'.$color.'" href="'.html($link).'">'
						.html($all_distrs->{$Did}{FULLNAME}).'</a>';
			if ( $distrs{$Did} ) {
				$res .= ' &nbsp; (found in <i>'.html($distrs{$Did}).'</i> linked by '.$soname.')';
			}
			$res .= '</li>'.$nl;
		}
		$res .= '</ul>'.$nl;

		return $res;
	};
	
	my $fullname = $Iname.($ver?'@'.$ver:'');
	my $approved = $Analysis::approved_libs->{$soname};
	
	my $id = 0;
	
	$res .= '<ul>';
	foreach my $com ( sort keys %{$int_obj->{COMPONENTS}} ) {
		$id++;
		my $com_obj = $Analysis::app_components->{$com};
		
		my (%present, %missing, %unknown);
		if ( 0 ) { # We've decided not to show the distribution list here, because its purpose isn't clear for user.
		foreach my $Did ( keys %$Analysis::all_distrs ) {
			my $tmp = $com_obj->{$Did}{FOUND_INTS}{$fullname};
			if ( $tmp ) {
				$tmp =~ s/\|.*//;
				my $lib_soname = $tmp;
				
				my $lib_obj = $com_obj->{$Did}{LIBS}{$lib_soname};
				if ( $lib_obj && $lib_obj->{VIA} ) { $lib_soname = $lib_obj->{VIA}; }
				
				if ( $lib_soname eq $soname ) {
					$present{$Did} = ($lib_soname eq $tmp ? "" : $tmp);
				}
				# else: from another lib
			}
			elsif( $tmp = $com_obj->{$Did}{MISSING_INTS}{$fullname} ) {
				$tmp =~ s/\|.*//;
				my $lib_soname = $tmp;
				
				my $lib_obj = $com_obj->{$Did}{LIBS}{$lib_soname};
				if ( $lib_obj && $lib_obj->{VIA} ) { $lib_soname = $lib_obj->{VIA}; }
				
				if ( $lib_soname eq $soname ) {
					if ( $approved ) {
						$missing{$Did} = ($lib_soname eq $tmp ? "" : $tmp);
					} else {
						$unknown{$Did} = ($lib_soname eq $tmp ? "" : $tmp);
					}
				}
				# else: from another lib
			}
			else {
				# Non-linked, ignore it.
			}
		}
		}
		
		my ($disp_name, $disp_path) = parse_app_component($com);
		$res .= '<li>';
		my $div_id = 'd'.$id;
		my $com_link = href(tab_link('com').'&com='.$com);
		#$res .= '[<a href="#" onclick="toggle(\''.$div_id.'\',this);return false">+</a>]&nbsp;';
		$res .= '<a href="'.html($com_link).'"'
			.' title="'.html($disp_path).'"><b>'.html($disp_name).'</b></a>';
		if ( 0 ) { # We've decided not to show the distribution list here, because its purpose isn't clear for user.
			$res .= '<div id="'.$div_id.'" style="padding:5pt;padding-bottom:10pt;display:none">';
			if ( %present ) {
				$res .= "<p>The interface is present in:</p>";
				$res .= distr_list('green', $com, %present);
			}
			if ( %missing ) {
				$res .= "<p>The interface is missing in:</p>";
				$res .= distr_list('red', $com, %missing);
			}
			if ( %unknown ) {
				$res .= "<p>Presence of the interface is unknown for:</p>";
				$res .= distr_list('#907000', $com, %unknown); # yellow
			}
			$res .= '</div>';
		}
		$res .= '</li>';
	}
	$res .= '</ul>';
	
	$res .= '<hr />'.$nl;
	
	return $res;
} # end of  int_details
#-----------------------------------------------------------------------

sub int_distr_list {
	my ($Iname, $ver, $distr_sel, $filt_com, $filt_lib) = @_;
	$ver ||= "";
	
	my $nl = "\n";
	my $res = "";
	
	my $ints = ints_data($distr_sel, $filt_com, $filt_lib);
	
	# Title
	my $unmangled = $Analysis::unmangled->{$Iname} || $Iname;
	# Reference to the LSB Navigator:
	my $navi_link = href("http://www.linuxfoundation.org/navigator/browse/int_single.php"
			."?cmd=list-by-name&Iname=".$Iname);
	$res .= '<h1>Interface: <a href="'.html($navi_link).'" target="_blank">'.html($unmangled).'</a>'.($ver ? ' ('.html($ver).')' : '').'</h1>';
	
	$res .= '<hr />'.$nl;
	
	# Presence in Distributions
	$res .= '<h2>Presence in Distributions</h2>';
	
	$res .= int_distr_list_sub($ints, $Iname, $ver, $distr_sel);
	
	$res .= '<hr />'.$nl;
}

sub int_distr_list_sub {
	my ($ints, $Iname, $ver, $distr_sel) = @_;
	$ver ||= "";
	
	my $nl = "\n";
	my $res = "";
	
	my $int = $ints->{$Iname}{$ver}
		or return "Error: There is no such interface: '$Iname' '$ver'";
	
	my $unapproved = $int->{UNAPPROVED};
	
	my %present = ();
	my %missing = ();
	foreach my $Did ( grep {$distr_sel->{$_}} keys %$distr_sel ) {
		my $distr_name = $Analysis::all_distrs->{$Did}{FULLNAME};
		if ( $int->{DISTRS}{$Did} ) {
			$present{$distr_name} = $Did;
		} else {
			$missing{$distr_name} = $Did;
		}
	}
	
	if ( $unapproved ) {
		$res .= '<p><i>No information.</i></p>';
		return $res;
	}
	
	my $filtered = scalar grep {!$_} values %$distr_sel;
	
	my $total_num = scalar grep {$_} values %$distr_sel;
	my $present_num = scalar keys %present;
	my $missing_num = scalar keys %missing;
	
	if ( $present_num > 0 ) {
		$res .= '<p>';
		$res .= 'Present in <span style="color:green">'.$present_num.'</span>';
		$res .= ' and missing in <span style="color:red">'.$missing_num.'</span>';
		if ( $filtered ) {
			$res .= ' of '.$total_num.' selected distributions.';
		} else {
			$res .= ' of total '.$total_num.' distributions.';
		}
		$res .= '</p>'.$nl;
	} else {
		$res .= '<p>This interface is missing in all the '.$total_num;
		if ( $filtered ) {
			$res .= ' selected distributions.</p>'.$nl;
		} else {
			$res .= ' analyzed distributions.</p>'.$nl;
		}
	}
	
	if ( %present ) {
		$res .= '<h3>Present in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %present ) {
			my $Did = $present{$distr_name};
			
			my $soname_linked = "";
			{ # Do not delete this block!
				last;
				foreach my $lib_with_path ( split /,\s*/, $int->{DISTRS}{$Did} ) {
					if ( $soname_linked ) { $soname_linked .= ', '; }
					if ( $lib_with_path =~ /^(.*)\/([^\/]+)$/ ) { # Link to lib info
						my ($path, $soname) = ($1, $2);
						my $lib_link = href( tab_link('libs').'&lib='.$soname );
						$soname_linked .= $path.'/'.'<a style="color:blue" href="'.html($lib_link).'">'.html($soname).'</a>';
					} else {
						$soname_linked .= $lib_with_path;
					}
				}
			}

			my $distr_link = href( tab_link('distrs').'&id='.$Did );
						
			$res .= '<li><a style="color:green" href="'.html($distr_link).'">'
						.html($distr_name).'</a>';
			#$res .= ' &nbsp; ('.$soname_linked.')';
			$res .= '</li>';
		}
		$res .= '</ul>';
	}
	
	if ( %missing ) {
		$res .= '<h3>Missing in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %missing ) {
			my $Did = $missing{$distr_name};
			
			my $distr_link = href( tab_link('distrs').'&id='.$Did );
			
			$res .= '<li><a style="color:red" href="'.html($distr_link).'">'
						.html($distr_name).'</a></li>';
		}
		$res .= '</ul>';
	}
	
	return $res;
} # end of  int_distr_list

#-----------------------------------------------------------------------

sub C_tab {
	my ($selected_lsb_ver) = @_;
	
	my $res = "";
	my $nl = "\n";

	#my $distr_sel = get_distr_sel();
	my $distr_sel = $Analysis::all_distrs;

	my $app_components = $Analysis::app_components;
	foreach my $com_obj ( values %$app_components ) {
		Analysis::com_data( $com_obj ); # TODO: use distr_sel
		Analysis::com_data_lsb( $com_obj, $selected_lsb_ver );
	}
	
	# Check if the table is not empty
	if ( !%$app_components ) {
		$res .= no_binary_div();
		return $res;
	}
	
	##
	
	my $link = href( tab_link('com') );
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[

function popup_win( link )
{
	var w = window.open( link, 'details',
			'toolbar=0,menubar=0,scrollbars=1,height=750,width=850,dependent=yes,dialog=yes'
		);
	w.moveTo(100,0);
	w.focus();

	return false;
}
// ]]>
</script>
HEREDOC

	$res .= <<HEREDOC;
<p>The table below lists all the binary components of your Application with their individual compatibility statuses.</p>
HEREDOC

	local *_pl = sub {
		my ($num, $mid, $singular, $plural) = @_;
		
		return '<b>'.$num.'</b> '.($mid ? $mid.' ' : '').($num > 1 ? $plural : $singular);
	};
	
	local *refer = sub {
		my ($com, $text, $tab, $add, $color) = @_;
		
		my $link = href(tab_link($tab).'&nofilt=1&filt_com='.href($com));
		if ( $add ) { $link .= href($add); }
		return '<a href="'.html(href($link)).'"'
			.($color?' style="color:'.html($color).'"':'')
			.'>'.$text.'</a>';
	};
	
	local *add_br = sub {
		$_[0] .= '<br/>' if $_[0];
	};

	$res .= '<table id="i_table" class="wikitable" width="100%">'.$nl;
	
	$res .= '<tr>'
			.'<th>Component</th>'
			.'<th>Distribution<br/>Compatibility</th>'
			.'<th>Missing<br/>Libraries</th>'
			.'<th>Missing<br/>Interfaces</th>'
			.'<th>Unneeded<br/>Libraries</th>'
			.'<th>LSB&nbsp;'.$selected_lsb_ver.'<br/>Compatibility</th>'
			.'<th>Comments</th>'
		.'</tr>'.$nl;
	
	foreach my $com ( sort keys %$app_components ) {
		my $com_obj = $app_components->{$com};
		
		$res .= '<tr>'.$nl;
		
		# Component
		$res .= '<td class="indentwrap">';
		my ($disp_name, $disp_path) = parse_app_component($com);
		$res .= '<span title="'.html($disp_path).'">'.html($disp_name).'</span>';
		$res .= '</td>'.$nl;
		
		# Distribution Compatibility
		my $all = scalar keys %$Analysis::all_distrs;
		
		my ($missing, $unknown, $ok) = (0) x 4;
		foreach my $Did ( keys %$Analysis::all_distrs ) {
			Analysis::com_distr_status( $com_obj, $Did );
			my $dstatus = $com_obj->{$Did}{STATUS};
			if ( $dstatus eq 'MISSING' || $dstatus eq 'UNNEEDED' ) { $missing++ }
			elsif ( $dstatus eq 'UNKNOWN' || $dstatus eq '' ) { $unknown++ }
			elsif ( $dstatus eq 'OK' ) { $ok++ }
		}
		my $bgcolor = ( $ok + $unknown == 0 ? '#FFCCCC' : $ok == $all ? '#CCFFCC' : '#FEFECC' );
		my $info_link = href( tab_link('com').'&com='.$com );
		my $info_onclick = 'onclick="return popup_win(\''.html($info_link).'\')"';
		$res .= '<td style="text-align:right;background:'.$bgcolor.'">';
		if ( $ok == 0 && $unknown > 0 ) {
			$res .= '<span style="white-space:nowrap">'.$unknown.' of '.$all.' unknown';
			$res .= ' (<a href="'.html($info_link).'" '.$info_onclick.'>list&hellip;</a>)';
			$res .= '</span>';
		} else {
			$res .= '<span style="white-space:nowrap">'.$ok.' of '.$all.'';
			$res .= ' (<a href="'.html($info_link).'" '.$info_onclick.'>list&hellip;</a>)';
			$res .= '</span>';
			if ( $unknown ) { $res .= ',<br/>+'.$unknown.' unknown'; }
		}
		$res .= '</td>'.$nl;
		
		# Missing Libraries
		$all = @{$com_obj->{NEEDED_LIBS}};
		my $x = 0; my $un = 0;
		foreach my $status ( values %{$com_obj->{LIB_STATUS}} ) {
			if ( $status eq 'MISSING' || $status eq 'UNNEEDED' ) { $x++ }
			elsif ( $status eq 'UNKNOWN' ) { $un++ }
		}
		$bgcolor = ( $x + $un == 0 ? '#CCFFCC' : $x == 0 ? '#FEFECC' : '#FFCCCC' );
			$res .= '<td style="text-align:right;background:'.$bgcolor.';white-space:nowrap">';
		if ( $x + $un == 0 ) {
			$res .= refer($com, 'none', 'libs', '&nofilt=1' );
		} elsif ( $x == 0 ) {
			$res .= $un.' unknown ('.refer($com, '&hellip;', 'libs', '&filt_show=problems').')';
		} else {
			$res .= $x.' of '.$all.' ('.refer($com, '&hellip;', 'libs', '&filt_show=problems').')';
			if ( $un ) { $res .= ',<br/>+'.$un.' unknown'; }
		}
		$res .= '</td>'.$nl;
		
		# Missing Interfaces
		$x = 0; $un = 0;
		$all = keys %{$com_obj->{NEEDED_INTS}};
		foreach my $status ( values %{$com_obj->{NEEDED_INTS}} ) {
			if ( $status eq 'MISSING' || $status eq 'UNNEEDED' ) { $x++ }
			elsif ( $status eq 'UNKNOWN' ) { $un++ }
		}
		$bgcolor = ( $x + $un == 0 ? '#CCFFCC' : $x == 0 ? '#FEFECC' : '#FFCCCC' );
		$res .= '<td style="text-align:right;background:'.$bgcolor.';white-space:nowrap">';
		if ( $x + $un == 0 ) {
			$res .= refer($com, 'none', 'ints', '&nofilt=1' );
		} elsif ( $x == 0 ) {
			$res .= $un.' unknown ('.refer($com, '&hellip;', 'ints', '&filt_show=problems').')';
		}
		 else {
			$res .= $x.' of '.$all.' ('.refer($com, '&hellip;', 'ints', '&filt_show=problems').')';
			if ( $un ) { $res .= ',<br/>+'.$un.' unknown'; }
		}
		$res .= '</td>'.$nl;
		
		# Unneeded libraries
		$x = grep {$_} values %{$com_obj->{LIBS_UNNEEDED}};
		$res .= '<td style="text-align:center">';
		if ( $x ) {
			$res .= '<span style="color:red"><b>'.$x.'</b></span>&nbsp;';
			$res .= '('.refer($com, '&hellip;', 'libs', '&filt_show=unneeded' ).')';
		}
		$res .= '</td>'.$nl;
		
		## LSB status
		my $lsb_status = "";
		# LSB status for Libraries
		my $lib_lsb_status = $com_obj->{LSB_STATUS}{LIBS};
		my %statuses;
		@statuses{qw/NOT CANDIDATE FUTURE IN OPTIONAL DEPRECATED WITHDRAWN/} = (0) x 7; # initialize
		foreach my $status_obj ( values %$lib_lsb_status ) {
			$statuses{ $status_obj->{STATUS} }++;
		}
		if ( $statuses{'NOT'} || $statuses{'WITHDRAWN'} ) {
			my $x = $statuses{'NOT'} + $statuses{'WITHDRAWN'} + $statuses{'CANDIDATE'} + $statuses{'FUTURE'};
			my $text = _pl($x, 'non-LSB', 'library', 'libraries');
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'libs', '&filt_show=non-lsb', '#DD0000'); # Red
		}
		elsif ( $statuses{'CANDIDATE'} ) {
			my $cand_version = (sort map {$_->{STATUS_VER}} grep {$_->{STATUS} eq 'CANDIDATE'} values %$lib_lsb_status)[-1]; # max
			my $text = "";
			if ( $statuses{'FUTURE'} ) {
				my $x = $statuses{'FUTURE'} + $statuses{'CANDIDATE'};
				$text = _pl($x, '', 'library', 'libraries')
						.' are candidates for LSB&nbsp;'.$cand_version.' or later.';
			} else {
				my $x = $statuses{'CANDIDATE'};
				$text = _pl($x, '', 'library', 'libraries')
						.' are candidates for LSB&nbsp;'.$cand_version;
			}
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'libs', '&filt_show=non-lsb', '#907000'); # Yellow
		}
		elsif ( $statuses{'FUTURE'} ) {
			my $future_version = (sort map {$_->{STATUS_VER}} grep {$_->{STATUS} eq 'FUTURE'} values %$lib_lsb_status)[-1]; # max
			my $x = $statuses{'FUTURE'};
			my $text = _pl($x, '', 'library is', 'libraries are')
				.' since LSB&nbsp;'.$future_version;
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'libs', '&filt_show=non-lsb', '#907000'); # Yellow
		}
		if ( $statuses{'DEPRECATED'} ) {
			my $x = $statuses{'DEPRECATED'};
			my $text = _pl($x, 'deprecated', 'library', 'libraries');
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'libs', '&filt_show=deprecated', '#907000'); # Yellow
		}
		# LSB status for Interfaces
		my $int_lsb_status = $com_obj->{LSB_STATUS}{INTS};
		@statuses{qw/NOT CANDIDATE FUTURE IN OPTIONAL DEPRECATED WITHDRAWN/} = (0) x 7; # initialize
		foreach my $status_obj ( values %$int_lsb_status ) {
			$statuses{ $status_obj->{STATUS} }++;
		}
		if ( $statuses{'NOT'} || $statuses{'WITHDRAWN'} ) {
			my $x = $statuses{'NOT'} + $statuses{'WITHDRAWN'} + $statuses{'CANDIDATE'} + $statuses{'FUTURE'};
			my $text = _pl($x, 'non-LSB', 'interface', 'interfaces');
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'ints', '&filt_show=non-lsb', '#DD0000'); # Red
		}
		elsif ( $statuses{'CANDIDATE'} ) {
			my $cand_version = (sort map {$_->{STATUS_VER}} grep {$_->{STATUS} eq 'CANDIDATE'} values %$int_lsb_status)[-1]; # max
			my $text = "";
			if ( $statuses{'FUTURE'} ) {
				my $x = $statuses{'FUTURE'} + $statuses{'CANDIDATE'};
				$text = _pl($x, '', 'interface', 'interfaces')
						.' are candidates for LSB&nbsp;'.$cand_version.' or later.';
			} else {
				my $x = $statuses{'CANDIDATE'};
				$text = _pl($x, '', 'interface', 'interfaces')
						.' are candidates for LSB&nbsp;'.$cand_version;
			}
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'ints', '&filt_show=non-lsb', '#907000'); # Yellow
		}
		elsif ( $statuses{'FUTURE'} ) {
			my $future_version = (sort map {$_->{STATUS_VER}} grep {$_->{STATUS} eq 'FUTURE'} values %$int_lsb_status)[-1]; # max
			my $x = $statuses{'FUTURE'};
			my $text = _pl($x, '', 'interface is', 'interfaces are')
				.' since LSB&nbsp;'.$future_version;
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'ints', '&filt_show=non-lsb', '#907000'); # Yellow			
		}
		if ( $statuses{'DEPRECATED'} ) {
			my $x = $statuses{'DEPRECATED'};
			my $text = _pl($x, 'deprecated', 'interface', 'interfaces');
			add_br( $lsb_status );
			$lsb_status .= refer( $com, $text, 'ints', '&filt_show=deprecated', '#907000'); # Yellow
		}
		if ( !$lsb_status ) {
			$lsb_status = '<span style="color:green">OK</span>';
		}
		$res .= '<td style="text-align:center">';
		$res .= $lsb_status;
		$res .= '</td>'.$nl;
		
		# Comments
		my $comments = "";
		if ( $com_obj->{NO_SHT_HASH} ) {
			$comments .= '<br/>' if $comments;
			$comments .= '<span style="white-space:nowrap">No .hash section!';
			$comments .= '&nbsp;<a href="" onclick="javascript:show_hide_tooltip(this, \'com_no_hash_section\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a></span>';
		}
		
		Analysis::com_common_data($com_obj);
		if ( $com_obj->{COMMON} ) {
			$comments .= '<br/>' if $comments;
			$comments .= '<span style="white-space:nowrap">A common library</span>';
			$comments .= '<span style="white-space:nowrap">';
			my $x = scalar grep { $distr_sel->{$_} } keys %{$com_obj->{COMMON_DISTRS}};
			my $all = scalar grep { $distr_sel->{$_} } keys %$distr_sel;
			my $link = href( tab_link('com').'&com='.$com.'&common=1' );
			my $onclick = 'onclick="return popup_win(\''.html($link).'\')"';
			$comments .= ' (<a href="'.html($link).'" '.$onclick.'>'.$x.' of '.$all.'</a>)';
			# Tooltip:
			$comments .= ' <a href="" onclick="javascript:show_hide_tooltip(this, \'com_common_lib\', event); return false;"><img src="images/question.png" width="16" height="16" alt="?" /></a>';
			$comments .= '</span>';
		}
		$res .= '<td align="right">';
		$res .= $comments;
		$res .= '</td>'.$nl;
		
		$res .= '</tr>'.$nl;
	}
	
	$res .= '</table>'.$nl;
	
	return $res;
}

# Contents of a details popup window for Components table
sub com_details {
	my ($com) = @_;

	my $nl = "\n";
	my $res = "";

	my $com_obj = $Analysis::app_components->{$com};
	if ( !$com_obj ) {
		$res .= "Error: there is no such component: '".html($com)."'";
		return $res;
	}
	Analysis::com_data( $com_obj );
	
	my ($disp_name, $disp_path) = parse_app_component($com);
	
	$res .= '<h1>Application&nbsp;Component<br/>'.html($disp_name).'</h1>'.$nl;
	
	if ( $disp_path ne $disp_name ) {
		$res .= '<p><b>Full path:</b> '.html($disp_path).'</p>';
	}
	
	$res .= '<hr/>'.$nl;
	
	if ( $com_obj->{NO_SHT_HASH} ) {
		$res .= '<h2>No \'.hash\' section!</h2>'.$nl;
		$res .= '<p>This application component has no <b>.hash</b> section. It would fail to run on older distributions (e.g., RHEL4).</p>'.$nl;
		$res .= '<p>Which section is present in the binary is determined by the linker switch \'--hash-style\' at compile time.</p>'.$nl;
		$res .= '<hr/>';
	}
	
	$res .= '<h2>Distribution Compatibility</h2>'.$nl;
	$res .= '<p>The table below shows the compatibility information for this application component.</p>'.$nl;
	$res .= '<table style="border-collapse:collapse" border="1">'.$nl;
	$res .= '<tr>'
			.'<th rowspan="2">Distribution</th>'
			.'<th colspan="2">Libraries</th>'
			.'<th colspan="2">Interfaces</th>'
			.( $com_obj->{NO_SHT_HASH} ? '<th rowspan="2">Comments</th>' : '' )
		.'</tr>'.$nl;
	$res .= '<tr>'
			.'<th>Missing</th>'
			.'<th>Unknown</th>'
			.'<th>Missing</th>'
			.'<th>Unknown</th>'
		.'</tr>'.$nl;
	my $all_distrs = $Analysis::all_distrs;
	my @distrs = sort {lc $all_distrs->{$a}{FULLNAME} cmp lc $all_distrs->{$b}{FULLNAME}} keys %$all_distrs;
	foreach my $Did ( @distrs ) {
		Analysis::com_distr_status( $com_obj, $Did );
		
		my $status = $com_obj->{$Did}{STATUS};
		my $bgcolor = "";
		if ( $status eq 'OK' ) {
			$bgcolor = "#ccffcc"; # green
		} elsif ( $status eq 'UNNEEDED' ) {
			$bgcolor = "#fefecc"; # yellow
		} elsif ( $status eq 'UNKNOWN' ) {
			$bgcolor = "#fefecc"; # yellow
		} elsif ( $status eq 'MISSING' ) {
			$bgcolor = "#ffdddd"; # red
		}
		my $link = href(tab_link('com').'&com='.$com.'&distr='.$Did);
		
		# Component Info
		my ($missing_libs, $unneeded_libs, $unknown_libs, $missing_ints, $unknown_ints) = (0) x 5;
		foreach my $req_soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			my $val = $com_obj->{$Did}{LIBS}{$req_soname}{STATUS};
			if ( $val eq 'MISSING' ) { $missing_libs++; }
			if ( $val eq 'UNNEEDED' ) { $unneeded_libs++; }
			if ( $val eq 'UNKNOWN' ) { $unknown_libs++; }
		}
		foreach my $val ( values %{$com_obj->{$Did}{INT_STATUS}} ) {
			if ( $val eq 'MISSING' ) { $missing_ints++ }
			if ( $val eq 'UNKNOWN' ) { $unknown_ints++; }
		}
		
		$res .= '<tr style="background:'.html($bgcolor).'">';
		$res .= '<td>';
		$res .= '<a href="'.html($link).'">'.html($all_distrs->{$Did}{FULLNAME}).'</a>';
		$res .= '</td>';
		
		$res .= '<td>';
		if ( $missing_libs ) {
			$res .= '<span style="color:red"><b>'.$missing_libs.'</b></span>';
			$res .= ' (+<span style="color:red"><b>'.$unneeded_libs.'</b></span> unneeded)' if $unneeded_libs;
		} elsif ( $unneeded_libs ) {
			$res .= '<span style="color:#907000">';
			$res .= '<b>'.$unneeded_libs.'</b> unneeded';
			$res .= '</span>';
		}
		$res .= '</td>';
		$res .= '<td><span style="color:#907000"><b>'.($unknown_libs or "").'</b></span></td>';
		$res .= '<td><span style="color:red"><b>'.($missing_ints or "").'</b></span></td>';
		$res .= '<td><span style="color:#907000"><b>'.($unknown_ints or "").'</b></span></td>';
		
		if ( $com_obj->{NO_SHT_HASH} ){
			$res .= '<td>';
			if ( $all_distrs->{$Did}{NO_GNU_HASH} ) {
				$res .= 'No .gnu.hash support';
			}
			$res .= '</td>';
		}
		
		$res .= '</tr>'.$nl;
	}
	$res .= '</table>'.$nl;
	
	$res .= '<hr />'.$nl;

	return $res;
}


# Contents of a details popup window for common lib comment at the Components tab
sub com_common_details {
	my ($com) = @_;

	my $nl = "\n";
	my $res = "";
	
	my $distr_sel = $Analysis::all_distrs;
	
	my $com_obj = $Analysis::app_components->{$com};
	Analysis::com_common_data($com_obj); # TODO: use distr_sel
	
	if ( !$com_obj ) {
		$res .= "<p>Error: there is no such component: '".html($com)."'</p>";
		return $res;
	}
	
	if ( !$com_obj->{COMMON} ) {
		$res .= "<p>No information.</p>";
		return $res;
	}

	my $link = href( 'http://www.linuxfoundation.org/navigator/browse/rawlib.php'
			.'?cmd=list-byrunname&RLrunname='.$com_obj->{SONAME} );
	$res .= '<h1><a href="'.html($link).'" target="_blank">'.html($com_obj->{SONAME}).'</a></h1>'.$nl;

	my %present = ();
	my %missing = ();
	foreach my $Did ( grep {$distr_sel->{$_}} keys %{$com_obj->{COMMON_DISTRS}} ) {
		my $distr_name = $Analysis::all_distrs->{$Did}{FULLNAME};
		$present{$distr_name} = $Did;
	}
	foreach my $Did ( grep {$distr_sel->{$_}} keys %$distr_sel ) {
		my $distr_name = $Analysis::all_distrs->{$Did}{FULLNAME};
		next if $present{$distr_name};
		$missing{$distr_name} = $Did;
	}

	# Explanation
	$res .= '<p>This library is provided by some distributions. Depending on which distributions provide this library (see below), you can consider excluding this library from the application package (making it an external dependency) thus reducing your package size.</p>'.$nl;

	$res .= '<hr/>'.$nl;
	
	$res .= '<h2><b>'.html($com_obj->{SONAME}).'</b> presence in distributions:</h2>';
	
	$res .= '<p>';
	$res .= 'Present in <span style="color:green">'.(keys %present).'</span>';
	$res .= ' and missing in <span style="color:red">'.(keys %missing).'</span>';
	$res .= ' of '.(grep {$distr_sel->{$_}} keys %$distr_sel).' distributions.';
	$res .= '</p>'.$nl;
	
	if ( %present ) {
		$res .= '<h3>Present in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %present ) {
			my $Did = $present{$distr_name};
			
			my $distr_link = href( tab_link('distrs').'&id='.$Did );
			
			$res .= '<li><a style="color:green" href="'.html($distr_link).'">'
						.html($distr_name).'</a></li>';
		}
		$res .= '</ul>';
	}

	if ( %missing ) {
		$res .= '<h3>Missing in:</h3>';
		$res .= '<ul>';
		foreach my $distr_name ( sort {lc $a cmp lc $b} keys %missing ) {
			my $Did = $missing{$distr_name};
			
			my $distr_link = href( tab_link('distrs').'&id='.$Did );
			
			$res .= '<li><a style="color:red" href="'.html($distr_link).'">'
						.html($distr_name).'</a></li>';
		}
		$res .= '</ul>';
	}
	
	$res .= '<hr />'.$nl;

	return $res;
}

sub com_distr_details {
	my ($com, $Did) = @_;

	my $nl = "\n";
	my $res = "";

	my $com_obj = $Analysis::app_components->{$com};
	if ( !$com_obj ) {
		$res .= "<p>Error: there is no such component: '".html($com)."'</p>";
		return $res;
	}
	my $distr = $Analysis::all_distrs->{$Did};
	if ( !$distr ) {
		$res .= "<p>Error: there is no such distribution: '".html($Did)."'</p>";
		return $res;
	}
	
	Analysis::com_distr_data( $com_obj, $Did, $com_obj->{LIBS_UNNEEDED}, $com_obj->{$Did}{NEEDED_LIBS_IMPL} );
	
	$res .= <<HEREDOC;
<script type="text/javascript" language="javascript">
// <![CDATA[
function toggle( id, obj )
{
	var v = document.getElementById( id );
	if ( v ) {
		if ( v.style.display == 'none' ) {
			v.style.display = '';
			if ( obj ) obj.innerHTML = '&ndash;';
		} else {
			v.style.display = 'none';
			if ( obj ) obj.innerHTML = '+';
		}
	}
	return false;
}
// ]]>
</script>
HEREDOC
	
	my ($disp_name, $disp_path) = parse_app_component($com);
	# Header
	$res .= '<h1>Application&nbsp;Component<br/><a href="'.html(href(tab_link('com').'&com='.$com)).'" title="'.html($disp_path).'">'.html($disp_name).'</a>'
			.' on&nbsp;<a href="'.html(href(tab_link('distrs').'&id='.$Did)).'" style="white-space:nowrap">'.html($distr->{FULLNAME}).'</a></h1>'.$nl;
	if ( $disp_path ne $disp_name ) {
		$res .= '<p><b>Full path:</b> '.html($disp_path).'</p>';
	}
	$res .= '<hr />'.$nl;
	
	if ( $com_obj->{NO_SHT_HASH} && $distr->{NO_GNU_HASH} ) {
		$res .= '<h2>Section .gnu.hash is not supported by the distribution</h2>'.$nl;
		$res .= '<p>Binaries targeting this distribution must use the \'.hash\' section instead.</p>'.$nl;
		$res .= '<hr />'.$nl;
	}
	
	# Gather data about libraries and interface linkage
	my $libs = {}; my $order = 0;
	foreach my $soname ( @{$com_obj->{NEEDED_LIBS}}, @{$com_obj->{$Did}{NEEDED_LIBS_IMPL}} ) {
		my $lib = $libs->{$soname} ||= { ORDER=>$order++ };
		
		my $com_lib = $com_obj->{$Did}{LIBS}{$soname};

		$lib->{STATUS} = $com_lib->{STATUS};

		$lib->{FILE} = $com_lib->{FILE} if $com_lib->{FOUND};
		if ( $com_lib->{VIA} ) {
			$lib->{VIA} = $com_lib->{VIA};
			push @{ $libs->{$com_lib->{VIA}}->{LINKED_LIBS} }, $soname;
			$lib->{IMPL} = 1;
		}
	}
	
	# Interfaces
	foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
		my $val;
		if ( $val = $com_obj->{$Did}{FOUND_INTS}{$fullname} ) {
			my ($soname, $ver) = split /\|/, $val;
			
			my $int = {STATUS=>'OK'};
			$int->{VER} = $ver if $ver;
			
			$lib->{INTERFACES}{$fullname} = $int;
		}
		elsif ( $val = $com_obj->{$Did}{MISSING_INTS}{$fullname} ) {
			my $soname = $val;
			my $lib = $libs->{$soname} ||= { STATUS=>'UNKNOWN', ORDER=>$order++, IMPL=>1 };
			$lib->{INTERFACES}{$fullname} = {STATUS=>'MISSING'};
		}
		else { # unlinked
			my $lib = $libs->{'NON-LINKED'} ||= { STATUS=>'MISSING', ORDER=>$order++ };
			$lib->{INTERFACES}{$fullname} = {STATUS=>'MISSING'};
		}
	}
	
	my $n = 0;
	foreach my $soname ( sort {$libs->{$a}{ORDER}<=>$libs->{$b}{ORDER}} keys %$libs ) {
		my $lib = $libs->{$soname};
		my $int_count = 0;
		my $int_mis_count = 0;
		if ( $lib->{INTERFACES} ) {
			$int_count = scalar keys %{$lib->{INTERFACES}};
			$int_mis_count = scalar grep {$lib->{INTERFACES}{$_}{STATUS} eq 'MISSING'} keys %{$lib->{INTERFACES}};
		}
		next if $lib->{IMPL} && !$int_count;
		
		my $int_linked_count = 0;
		if ( $lib->{LINKED_LIBS} ) {
			foreach my $soname ( @{$lib->{LINKED_LIBS}} ) {
				$int_linked_count += keys %{$libs->{$soname}{INTERFACES}};
			}
		}
		
		$n++;
		
		$res .= '<p>';
		
		my $block_id = 'b'.$n;
		if ( $int_count || $lib->{LINKED_LIBS} ) {
			$res .= '[<a href="#" onclick="toggle(\''.$block_id.'\', this);return false;">+</a>] ';
		} else {
			$res .= '[0] ';
		}
		
		my $color = undef;
		my $add = "";
		if ( $lib->{STATUS} eq 'OK' ) {
			$color = 'green';
			$add .= 'Runname: '.$lib->{FILE} if $lib->{FILE};
			# TODO: DISTR_COMPONENT
		}
		elsif ( $lib->{STATUS} eq 'UNKNOWN' ) {
			$color = '#907000'; # yellow
			if ( !$lib->{IMPL} || $lib->{VIA} ) {
				$add .= 'Status is unknown';
			}
		}
		elsif ( $lib->{STATUS} eq 'UNNEEDED' ) {
			$color = '#907000'; # yellow
			$add .= 'Missing library';
		}
		elsif ( $lib->{STATUS} eq 'MISSING' ) {
			$color = 'red';
			$add .= 'Missing library' if $soname ne 'NON-LINKED';
		}
		
		if ( $lib->{IMPL} ) {
			$add .= ', ' if $add;
			$add .= '<span style="color:red">Not in the NEEDED list</span>';
			if ( $lib->{VIA} ) {
				$add .= ', linked by '.html($lib->{VIA});
			}
		}
		
		$add .= ', ' if $add;
		if ( $int_count == 0 ) {
			# <yellow>0</yellow> interfaces
			$add .=  '<span style="color:#907000"><b>0</b></span> interfaces';
		}
		elsif ( $lib->{STATUS} eq 'MISSING' ) {
			# <red> N interfaces </red>
			$add .=  '<span style="color:red"><b>'.$int_count.'</b>'
				.' interface'.($int_count==1?'':'s').'</span>';
		}
		elsif ( $lib->{STATUS} eq 'UNKNOWN' ) {
			# <yellow> N interfaces unknown </yellow>
			$add .= '<span style="color:#907000"><b>'.$int_mis_count.'</b> interfaces - no information</span>';
		}
		else {
			# N interfaces
			$add .= '<b>'.$int_count.'</b> interface'.($int_count==1?'':'s');
			
			if ( $int_mis_count ) {
				$add .= ', ' if $add;
				# <red> M missing </red>
				$add .= '<span style="color:red"><b>'.$int_mis_count.'</b> missing</span>';
			}
			else {
				$add .= ', ' if $add;
				$add .= '0 missing';
			}
		}
		
		if ( $int_linked_count ) {
			$add .= ', ' if $add;
			$add .= ' +'.$int_linked_count.' interface'.($int_linked_count==1?'':'s').' in linked libs';
		}
		
		if ( $color ) {
			$res .= '<span style="color:'.$color.'"><b>'.html($soname).'</b></span>';
		} else {
			$res .= '<b>'.html($soname).'</b>';
		}
		$res .= ' &nbsp; ('.$add.')' if $add;
		$res .= '</p>'.$nl;
		
		$res .= '<div id="'.$block_id.'" style="display:none">'.$nl;
		# Interfaces of the library
		if ( $lib->{INTERFACES} ) {
			$res .= '<ul>'.$nl;
			foreach my $fullname ( sort {lc $a cmp lc $b} keys %{$lib->{INTERFACES}} ) {
				my $int = $lib->{INTERFACES}{$fullname};
				$res .= '<li>';
				my $color = "";
				my $add = "";
				if ( $int->{STATUS} eq 'MISSING' ) {
					if ( $lib->{STATUS} eq 'UNKNOWN' ) {
						$color = '#907000'; # Yellow
					} else {
						$color = 'red';
					}
				}
				elsif ( $int->{STATUS} eq 'OK' ) {
					$add .= $int->{VER} if $int->{VER}; # TODO: test it
				}
				
				my ($Iname, $ver) = split /\@/, $fullname;
				my $displayname = $Analysis::unmangled->{$Iname} || $Iname;
				$displayname .= ' @'.$ver if $ver;
				if ( $color ) {
					$res .= '<span style="color:'.$color.'">'.html($displayname).'</span>';
				} else {
					$res .= html($displayname);
				}
				$res .= ' &nbsp; ('.$add.')' if $add;
				
				# TODO: other versions?
				$res .= '</li>'.$nl;
			}
			$res .= '</ul>'.$nl;
		}
		# Interfaces of the library
		if ( $lib->{LINKED_LIBS} ) {
			$res .= '<ul>'.$nl;
			foreach my $soname ( @{$lib->{LINKED_LIBS}} ) {
				my $linked_ints_n = keys %{$libs->{$soname}{INTERFACES}};
				$res .= '<li>';
				$res .= 'Linked library <b>'.html($soname).'</b>';
				$res .= ' with <b>'.$linked_ints_n.'</b> interface'.($linked_ints_n==1?'':'s');
				$res .= ' (see below)';
				$res .= '</li>'.$nl;
			}
			$res .= '</ul>'.$nl;
		}
		$res .= '</div>'.$nl;
	}
	
	$res .= '<hr />'.$nl;
	
	return $res;
}
#-----------------------------------------------------------------------

sub ajax_wait {
	return <<HEREDOC;
<div id="ajax_loading" style="display: none; position: fixed; width: 13em; height: 5.8em; top: 40%; left: 50%; margin-top: -2.9em; margin-left: -7.5em; border: solid 1px #00007f; background-color: #e7f0ff; text-align: center; padding: 6px;"><div align="center">
Please wait&hellip;<br />
<img src="images/environment-spacer.gif" width="1" height="5" alt="" /><br />
<img src="images/ajax_progress.gif" width="16" height="16" alt="" /><br />
<img src="images/environment-spacer.gif" width="1" height="7" alt="" /><br />
<input type="button" name="ajax_abort" id="ajax_abort" value="Abort" style="border: solid 1px #00007f; background: #e0e0ff; color: blue; width: 6em;" onclick="javascript:onAbort();" />
</div></div>
HEREDOC

}

1; # return value
